/*		 ITI-FG  Kernel module free public release
 *
 *		     Copyright (C) 1996-2000 GOM mbH.
 *	         Rebenring 33, 38106 Braunschweig, GERMANY
 *		            All rights reserved
 *
 * $Id: itifgAM.c,v 7.0.2.1 2000/11/29 15:00:02 cvs Exp $
 *
 */

/*
 * Includes:
 */

#include "itifgOS.h"	/* os specific stuff */
#include "itifgExt.h"	/* offical driver interface */
#include "itifgReg.h"	/* register definitions ICR Regs */
#include "icpciReg.h"	/* register definitions IC-PCI */
#include "amvsReg.h"	/* register definitions AM-VS */
#include "amdigReg.h"	/* register definitions AM-DIG */
#include "ampcvReg.h"	/* register definitions PCVision */
#include "pcdigReg.h"	/* register definitions PC-DIG */
#include "itifgIntD.h"	/* driver internals */
#include "itifgIntF.h"	/* internal functions */

/*
 * Defines:
 */

#define LUT16_BANK_SIZE	0x100	/* 256 bytes per bank */

#define SET_LUTPG_BY_DEVICE(page) \
	if (IS_AMDG) \
	  { \
	    if (dg_set_lutpg (iti, page)) goto error; \
	  } \
	else \
	  { \
	    if (device == '0') \
	      { \
	        if (pcd_set_lut0pg (iti, page)) goto error; \
	      } \
	    else \
	      { \
	        if (pcd_set_lut1pg (iti, page)) goto error; \
	      } \
	  } \

/*
 * Globals:
 */

const struct vs_cam_t	/* default AM-VS camera structure
			   see amvsReg.h for description */

vs_def_cam = { /*---control register--------------------------------*/
  	       (VS_CON1_INIT & VS_TSEL_MASK) >> VS_TSEL_SHIFT,
	       (VS_CON1_INIT & VS_CLKSRC_MASK) >> VS_CLKSRC_SHIFT,
	       (VS_CON1_INIT & VS_LPFSEL_MASK) >> VS_LPFSEL_SHIFT,
	       (VS_CON1_INIT & VS_ISCALE_MASK) >> VS_ISCALE_SHIFT,
	       (VS_CON1_INIT & VS_SMODE_MASK) >> VS_SMODE_SHIFT,
	       /*---acquisition control register--------------------*/
	       (VS_CON2_INIT & VS_TSRC_MASK) >> VS_TSRC_SHIFT,
	       (VS_CON2_INIT & VS_LENPOL_MASK) >> VS_LENPOL_SHIFT,
	       (VS_CON2_INIT & VS_FLDPOL_MASK) >> VS_FLDPOL_SHIFT,
	       (VS_CON2_INIT & VS_FENPOL_MASK) >> VS_FENPOL_SHIFT,
	       (VS_CON2_INIT & VS_PCLKPOL_MASK) >> VS_PCLKPOL_SHIFT,
	       (VS_CON2_INIT & VS_ECLKEN_MASK) >> VS_ECLKEN_SHIFT,
	       (VS_CON2_INIT & VS_FLDSTEN_MASK) >> VS_FLDSTEN_SHIFT,
	       (VS_CON2_INIT & VS_VSTEN_MASK) >> VS_VSTEN_SHIFT,
	       /*---trigger control register------------------------*/
	       (VS_TRIGCON_INIT & VS_TRIGEN_MASK) >> VS_TRIGEN_SHIFT,
	       (VS_TRIGCON_INIT & VS_TRIGSRC_MASK) >> VS_TRIGSRC_SHIFT,
	       (VS_TRIGCON_INIT & VS_FMODE_MASK) >> VS_FMODE_SHIFT,
	       (VS_TRIGCON_INIT & VS_TMODE_MASK) >> VS_TMODE_SHIFT,
	       (VS_TRIGCON_INIT & VS_STRBDLY_MASK) >> VS_STRBDLY_SHIFT,
	       /*---geometry and delay information------------------*/
	       VS_HOFF_INIT,
	       VS_HSZ_INIT,
	       VS_VOFF_INIT,
	       VS_VSZ_INIT,
	       VS_FROFF_INIT,
	       /*---clamp control register--------------------------*/
	       (VS_CPCON_INIT & VS_CPPOS_MASK) >> VS_CPPOS_SHIFT,
	       (VS_CPCON_INIT & VS_CPEN_MASK) >> VS_CPEN_SHIFT,
	       (VS_CPCON_INIT & VS_CPSRC_MASK) >> VS_CPSRC_SHIFT,
	       (VS_CPCON_INIT & VS_STRBSEL_MASK) >> VS_STRBSEL_SHIFT,
	       (VS_CPCON_INIT & VS_STRBPOL_MASK) >> VS_STRBPOL_SHIFT,
	       (VS_CPCON_INIT & VS_CPST_MASK) >> VS_CPST_SHIFT,
	       /*---reference voltages------------------------------*/
	       VS_NREF_INIT,
	       VS_PREF_INIT
	     };

struct dg_cam_t	/* default AM-DIG camera structure
		   see amdigReg.h for description */

dg_def_cam = { /*---camera clock control----------------------------*/
  	       (DG_CLKCON_INIT & DG_FSEL_MASK) >> DG_FSEL_SHIFT,
	       (DG_CLKCON_INIT & DG_CLKDIV_MASK) >> DG_CLKDIV_SHIFT,
	       (DG_CLKCON_INIT & DG_XILLEN_MASK) >> DG_XILLEN_SHIFT,
	       (DG_CLKCON_INIT & DG_MC_MASK) >> DG_MC_SHIFT,
	       /*---camera timing control---------------------------*/
	       (DG_CTCON_INIT & DG_PCLKPOL_MASK) >> DG_PCLKPOL_SHIFT,
	       (DG_CTCON_INIT & DG_LENPOL_MASK) >> DG_LENPOL_SHIFT,
	       (DG_CTCON_INIT & DG_FENPOL_MASK) >> DG_FENPOL_SHIFT,
	       (DG_CTCON_INIT & DG_FLDPOL_MASK) >> DG_FLDPOL_SHIFT,
	       (DG_CTCON_INIT & DG_SMODE_MASK) >> DG_SMODE_SHIFT,
	       (DG_CTCON_INIT & DG_ILMODE_MASK) >> DG_ILMODE_SHIFT,
	       (DG_CTCON_INIT & DG_LENSM_MASK) >> DG_LENSM_SHIFT,
	       (DG_CTCON_INIT & DG_NOLMISS_MASK) >> DG_NOLMISS_SHIFT,
	       (DG_CTCON_INIT & DG_LVAR_MASK) >> DG_LVAR_SHIFT,
	       /*---external sync control and time------------------*/
	       (DG_EXCON_INIT & DG_EXCLK_MASK) >> DG_EXCLK_SHIFT,
	       (DG_EXCON_INIT & DG_EXTMD_MASK) >> DG_EXTMD_SHIFT,
	       (DG_EXCON_INIT & DG_EXSEN_MASK) >> DG_EXSEN_SHIFT,
	       (DG_EXCON_INIT & DG_EXSPOL_MASK) >> DG_EXSPOL_SHIFT,
	       (DG_EXCON_INIT & DG_PRIEN_MASK) >> DG_PRIEN_SHIFT,
	       (DG_EXCON_INIT & DG_PRIP_MASK) >> DG_PRIP_SHIFT,
	       DG_EXTIME_INIT,
	       /*---lut control and page select---------------------*/
	       (DG_LUTCON_INIT & DG_PSIZE_MASK) >> DG_PSIZE_SHIFT,	       
	       (DG_LUTCON_INIT & DG_ILPIXEL_MASK) >> DG_ILPIXEL_SHIFT,	       
	       (DG_LUTCON_INIT & DG_MUXA_MASK) >> DG_MUXA_SHIFT,	       
	       (DG_LUTCON_INIT & DG_MUXB_MASK) >> DG_MUXB_SHIFT,	       
	       (DG_LUTCON_INIT & DG_MUXC_MASK) >> DG_MUXC_SHIFT,	       
	       /*---geometry information----------------------------*/
	       DG_HOFF_INIT,
	       DG_HACT_INIT,
	       DG_VOFF_INIT,
	       DG_VACT_INIT,
	       /*---trigger control---------------------------------*/
	       (DG_TRIGCON_INIT & DG_TRIGEN_MASK) >> DG_TRIGEN_SHIFT,
	       (DG_TRIGCON_INIT & DG_TRIGEN_MASK) >> DG_TRIGEN_SHIFT,
	     };

const struct pv_cam_t	/* default PCVision camera structure
			   see ampcvReg.h for description */

pv_def_cam = { /* line registers */
	       PV_HTOTAL_INIT,
	       PV_HSYNCPW_INIT,
	       PV_HEQUIPW_INIT,
	       PV_HSERRPW_INIT,
	       /* frame registers */
	       PV_VTOTAL_INIT,
	       PV_VSYNCPW_INIT,
	       PV_HS_ACT_LOW,	/* HSYNC/VSYNC are Active Low (invert) */
	       PV_VS_ACT_LOW,
	       PV_VGSTART_INIT,
	       PV_VGWIDTH_INIT,
	       /* window registers */
	       PV_HOFFSET_INIT,
	       PV_HACTIVE_INIT,
	       PV_VOFFSET_INIT,
	       PV_VACTIVE_INIT,
	       /* control registers */
	       PV_STRBDLY_INIT,
	       PV_LP_ACT_LOW,	/* LEN/FEN Active Low (invert) */
	       PV_FP_ACT_LOW,
	       PV_FROFF_INIT,
	       PV_DISABLE,	/* Frame Reset on Vsync is Disabled */
	       PV_DISABLE,	/* Frame Reset is Disabled */
	       PV_FR_ACT_LOW,	/* Frame Reset Pulse is Active Low */
	       PV_FRST_FROFF,	/* Frame Reset is True for FROFF Period */
	       PV_HSYNC_OUT,	/* Horizontal Sync Output */
	       PV_CHSYNC_DIS,	/* CHSYNC Output disabled */
	       PV_VSYNC_DIS,	/* VSYNC Output disabled*/
	       PV_CLKSRC_PLL,	/* PLL-Mode / Strip field from video */
	       PV_LPF_6_0,	/* Filter enabled (6Mhz cut off -3dB) */
	       PV_DISABLE,	/* Disable VariableScan Mode */
	       PV_SCALE_1_5,	/* Input Gain x 1.5 */
	       PV_CP_ACT_LOW,	/* Sample VariableScan Clock on Falling Edge */
	       PV_ILACE,	/* Interlaced */
	       PV_CAM_0,	/* Camera 0 */
	       PV_DISABLE,	/* Normal Field Shift Mode */
	       /* trigger/strobe registers */
	       PV_DISABLE,	/* Trigger Disabled */ 
	       PV_TRIG_SOFT,	/* Software Trigger from SW_TRIG register*/ 
	       PV_TP_ACT_LOW,	/* Extrenal Trigger Pulse is Active Low */ 
	       PV_STRBMD_FAST,	/* Fast Strobe */
	       PV_DISABLE,	/* Disable Strobe Output */
	       PV_SP_ACT_LOW,	/* Strobe Output is Active Low */ 
	       PV_DISABLE,	/* Disable Skip Field Mode */ 
	       PV_CCNT_INIT,
	       PV_CLMP_NO_INV,	/* HSYNC Active Low -> Falling Edge starts */
	       PV_CLMP_STRIP,	/* Clamp Pulse Comes from Sync Stripper */
	       /* pll registers */
	       PV_FBKDIV_INIT,
	       PV_REFDIV_INIT,
	       PV_VCO_INIT,
	       PV_PDA_INIT,
	       PV_LCNT_INIT,
	       /* reference voltage */
	       PV_NREF_INIT,
	       PV_PREF_INIT,
	     };

const struct pcd_cam_t	/* default PCDig camera structure
			   see pcdigReg.h for description */

pcd_def_cam = {
		/* PCD_CTCON_INIT */
#define PCD_CTCON_INIT		0x0000007D
		PCD_ENABLE,		/* LEN Input Enable */
		PCD_LEN_SYNC,		/* LEN Input Synchronous Sample */
		PCD_LEN_ACT_HIGH,	/* Rising Edge Begins New Line */
		PCD_ENABLE,		/* FEN Input Enable */
		PCD_FEN_ACT_HIGH,	/* Rising Edge Begins New Frame */
		PCD_ENABLE,		/* MULTI Input Enable */
		PCD_MULTI_ACT_HIGH,	/* MULTI Input is Active High */
		PCD_MULTI_FIELD,	/* Field Information at MULTI Input */
		/* PCD_OUTCON_INIT */
		PCD_DISABLE,		/* Timing Output Retime Disable */
		PCD_RESTIME_ACT_LOW,	/* Synchronize to Falling Edge */
		PCD_DISABLE,		/* Camera 3 Timing Output Disable */
		PCD_ENABLE,		/* External Sync Mode Enable */
		PCD_EXSYNC_ACT_HIGH,	/* EXSYNC goes high when triggered */
		PCD_EXSYNC_FREE_RUN,	/* External Sync Mode is FreeRunning */
		PCD_ENABLE,		/* Programmable Integration Enable */
		PCD_PRI_ACT_HIGH,	/* PRI Asserted High */
		0,			/* Camera Mode Control Value is 0 */
		PCD_CCLK_DISABLE,	/* CCLK Output to Camera Disable */
		PCD_CCLK_STATIC_HIGH,	/* CCLK Pin is High */
		/* PCD_EXSCON_INIT */
		PCD_EXSSEL_STRT_TO_MID,
			/* EXSYNC Starts at Begin and Stops at Mid */
		PCD_PRISEL_STRT_TO_MID,
			/* PRI Starts at Begin and Stops at Mid */
		PCD_PRIHOLD_NONE,
			/* Do not Extend the PRI Pulse */
		
		PCD_EXTIME_INIT,	/* double value */
		PCD_ROTIME_INIT,	/* double value */
		PCD_HOFF_INIT,
		PCD_HACT_INIT,
		PCD_VOFF_INIT,
		PCD_VACT_INIT,
		PCD_CCLKFREQ_INIT,	/* float value */
		/* PCD_LUTCON0_INIT */
		0,	/* LUT configuration is 8bit/256pages */
		0,	/* Select LUT page 0 */
		/* PCD_LUTCON1_INIT */
		0,	/* LUT configuration is 8bit/256pages */
		0,	/* Select LUT page 0 */
		/* PCD_TRIGCON_INIT */
		PCD_TRIG_ACT_LOW,
			/* Falling Edge/Low Level Causes Trigger 0 */
		PCD_TRIG_SRC_OPTO,
			/* Opto-Coupled Trigger 0 Input */
		PCD_DISABLE,
			/* Disable Trigger 0 Glitch Detector */
		PCD_TRIG_ACT_LOW,
			/* Falling Edge/Low Level Causes Trigger 1 */
		PCD_TRIG_SRC_OPTO,
			/* Opto-Coupled Trigger 1 Input */
		PCD_DISABLE,
			/* Disable Trigger 1 Glitch Detector */
		/* PCD_TRIGDIV_INIT */
		0,	/* Divide Harware Trigger by 1 */
		PCD_DIVCON_CLEAR,	/* Disable and Clear Counter */
		PCD_1CHAN_L2R	/* standard geometry */
	      };
	     
/*
 *  icp_get_modid
 */

u_char
icp_get_modid (itifg_t *iti)
{
  u_short tmp_val;
  
  tmp_val = iti_readw (iti->v_mod_base + ICP_MODID_OFFST);
  
  return ((tmp_val & ICP_TYPE_MASK) >> ICP_TYPE_SHIFT);
}

/*
 * iti_get_lut16
 */

int
iti_get_lut16 (itifg_t *iti, char device, union iti_lut16_t *get_lut)
{
  int count = 0, page = 0, bank = 0, banks = 0;
  u_short data = 0;
  
  u_char psize;

  extern int
    dg_write_ms (itifg_t*, u_short, u_short, u_char, u_char);
  extern u_short
    dg_read_mm (itifg_t*, u_char, u_short, u_short, u_short);
  
#if(0)
  if (IS_AMDG)
    {
      u_char mod_ver = dg_get_modver (iti);
      if (modver == DG_MODVER_8_TTL || modver == DG_MODVER_8_RS422)
	{
	  iti->errno = -EITI_NOSUPPORT;
	  return ERROR;
	}
    }
  if (get_lut == NULL) iti->errno = -EITI_POINTER;
#endif

  switch (MODULE_ID)
    {
    case ICP_AMDG:
      if (device != '0')
	{
	  iti->errno = -EITI_NOSUPPORT;
	  return ERROR;
	}
	psize = dg_get_psize (iti);
      break;
    case ITI_VMPCD:
      switch (device)
	{
	case '0':
	  psize = pcd_get_lut0psz (iti);
	  break;
	case '1':
	  psize = pcd_get_lut1psz (iti);
	  break;
	default:
	  iti->errno = -EITI_NOSUPPORT;
	  return ERROR;
	}
      break;
    default:
      iti->errno = -EITI_NOSUPPORT;
      return ERROR;
    }

  switch (psize)
    {
    case 16:
      banks = 1 << DG_PSIZE_16_SHIFT;
      break;
    case 14:
      if (MODULE_ID == ICP_AMDG)
	{
	  /* error at am-dig */
	  iti->errno = -EITI_NOSUPPORT;
	  return ERROR;
	}
      else
	{
	  /* 14bit at pc-dig */
	  banks = 1 << PCD_PSIZE_14_SHIFT;
	}
      break;
    case 12:
      banks = 1 << DG_PSIZE_12_SHIFT;
      break;
    case 10:
      banks = 1 << DG_PSIZE_10_SHIFT;
      break;
    case 8:
      banks = 1 << DG_PSIZE_8_SHIFT;
      break;
    }
  
  if (IS_AMDG)
    /* am-dig */
    page = dg_get_lutpg (iti);
  else
    {
      /* pc-dig */
      if (device == '0')
	/* lut 0 */
	page = pcd_get_lut0pg (iti);
      else
	/* lut 1 */
	page = pcd_get_lut1pg (iti);
      if (pcd_set_lutacc (iti, PCD_ENABLE)) return ERROR;
    }
  
  for (bank = 0; bank < banks; bank++)
    {
      if (IS_AMDG)
	if (dg_set_lutpg (iti, page + bank)) return ERROR;
      
      for (count = 0; count < LUT16_BANK_SIZE; count++)
	{
	  if (IS_AMDG)
	    data = dg_read_mm (iti,
			       DG_LUT16_OFFSET + count,
			       DG_LUT16_MASK,
			       DG_LUT16_MIN,
			       DG_LUT16_MAX);
	  else
	    {
	      if (device == '0')
		data = iti_readl (&iti->pcd_lut0_table
				  [(page + bank) * LUT16_BANK_SIZE + count]);
	      else
		data = iti_readl (&iti->pcd_lut1_table
				  [(page + bank) * LUT16_BANK_SIZE + count]);
	    }
	  get_lut->lut16data[bank * LUT16_BANK_SIZE + count] = data;
	}
    }
  
  if (IS_AMDG)
    {
      if (dg_set_lutpg (iti, page)) return ERROR;
    }
  else
    {
      if (pcd_set_lutacc (iti, PCD_DISABLE)) return ERROR;
    }

  return OK;
}

/*
 * iti_set_lut16
 */

int
iti_set_lut16 (itifg_t *iti, char device, union iti_lut16_t *set_lut)
{
  int count = 0, page = 0, bank = 0, banks = 0;
  u_short data = 0;

  u_char psize;

  extern int
     dg_write_ms (itifg_t*, u_short, u_short, u_char, u_char);
  extern int
    dg_write_mm (itifg_t*, u_short, u_short, u_short, u_short, u_short);

  if (IS_AMDG)
    {
      u_char mod_ver = dg_get_modver (iti);
      if (mod_ver == DG_MODVER_8_TTL || mod_ver == DG_MODVER_8_RS422)
	{
	  iti->errno = -EITI_NOSUPPORT;
	  return ERROR;
	}
    }
  if (set_lut == NULL)
    {
      iti->errno = -EITI_POINTER;
      return ERROR;
    }
  
  switch (MODULE_ID)
    {
    case ICP_AMDG:
      if (device != '0')
	{
	  iti->errno = -EITI_NOSUPPORT;
	  return ERROR;
	}
	psize = dg_get_psize (iti);
      break;
    case ITI_VMPCD:
      switch (device)
	{
	case '0':
	  psize = pcd_get_lut0psz (iti);
	  break;
	case '1':
	  psize = pcd_get_lut1psz (iti);
	  break;
	default:
	  iti->errno = -EITI_NOSUPPORT;
	  return ERROR;
	}
      break;
    default:
      iti->errno = -EITI_NOSUPPORT;
      return ERROR;
    }

  switch (psize)
    {
    case 16:
      banks = 1 << DG_PSIZE_16_SHIFT;
      break;
    case 14:
      if (MODULE_ID == ICP_AMDG)
	{
	  /* error at am-dig */
	  iti->errno = -EITI_NOSUPPORT;
	  return ERROR;
	}
      else
	{
	  /* 14bit at pc-dig */
	  banks = 1 << PCD_PSIZE_14_SHIFT;
	}
      break;
    case 12:
      banks = 1 << DG_PSIZE_12_SHIFT;
      break;
    case 10:
      banks = 1 << DG_PSIZE_10_SHIFT;
      break;
    case 8:
      banks = 1 << DG_PSIZE_8_SHIFT;
      break;
    }
  
  if (IS_AMDG)
    /* am-dig */
    page = dg_get_lutpg (iti);
  else
    {
      /* pc-dig */
      if (device == '0')
	/* lut 0 */
	page = pcd_get_lut0pg (iti);
      else
	/* lut 1 */
	page = pcd_get_lut1pg (iti);
      if (pcd_set_lutacc (iti, PCD_ENABLE)) return ERROR;
    }

  for (bank = 0; bank < banks; bank++)
    {
      if (IS_AMDG)
	if (dg_set_lutpg (iti, page + bank)) return ERROR;

      for (count = 0; count < LUT16_BANK_SIZE; count++)
	{
	  data = set_lut->lut16data[bank * LUT16_BANK_SIZE + count];
	  if (IS_AMDG)
	    {
	      if (dg_write_mm (iti,
			       DG_LUT16_OFFSET + count,
			       DG_LUT16_MASK,
			       DG_LUT16_MIN,
			       DG_LUT16_MAX,
			       data) < 0) return (ERROR);
	    }
	  else
	    {
	      if (device == '0')
		iti_writel (data, &iti->pcd_lut0_table
			    [(page + bank) * LUT16_BANK_SIZE + count]);
	      else
		iti_writel (data, &iti->pcd_lut1_table
			    [(page + bank) * LUT16_BANK_SIZE + count]);
	    }
	}
    }

  if (IS_AMDG)
    {
      if (dg_set_lutpg (iti, page)) return ERROR;
    }
  else
    {
      if (pcd_set_lutacc (iti, PCD_DISABLE)) return ERROR;
    }

  return OK;
}

/*
 * iti_lun_lut16
 */

int
iti_lin_lut16 (itifg_t *iti, char device, iti_lut16md_t mode)
{
  int count = 0, shift = 0, page = 0, pages = 0, banks = 0;
  u_short data = 0;

  u_char psize;

  union iti_lut16_t *lin_lut;  
  int lin_lut_size = 0;

  switch (MODULE_ID)
    {
    case ICP_AMDG:
      if (device != '0')
	{
	  iti->errno = -EITI_NOSUPPORT;
	  return ERROR;
	}
	psize = dg_get_psize (iti);
      break;
    case ITI_VMPCD:
      switch (device)
	{
	case '0':
	  psize = pcd_get_lut0psz (iti);
	  break;
	case '1':
	  psize = pcd_get_lut1psz (iti);
	  break;
	default:
	  iti->errno = -EITI_NOSUPPORT;
	  return ERROR;
	}
      break;
    default:
      iti->errno = -EITI_NOSUPPORT;
      return ERROR;
    }

  switch (psize)
    {
    case 16:
      shift = DG_PSIZE_16_SHIFT;
      pages = DG_PSIZE_16_MASK + 1;
      banks = 1 << DG_PSIZE_16_SHIFT;
      lin_lut_size = sizeof (lin_lut->lut16data);
      break;
    case 14:
      if (IS_AMDG)
	{
	  /* error at am-dig */
	  iti->errno = -EITI_NOSUPPORT;
	  return ERROR;
	}
      else
	{
	  /* 14bit at pc-dig */
	  shift = PCD_PSIZE_14_SHIFT;
	  pages = 1 << (8 - PCD_PSIZE_14_SHIFT);
	  banks = 1 << PCD_PSIZE_14_SHIFT;
	  lin_lut_size = sizeof (lin_lut->lut14data);
	}
      break;
    case 12:
      shift = DG_PSIZE_12_SHIFT;
      pages = 1 << (8 - DG_PSIZE_12_SHIFT);
      banks = 1 << (DG_PSIZE_12_SHIFT);
      lin_lut_size = sizeof (lin_lut->lut12data);
      break;
    case 10:
      shift = DG_PSIZE_10_SHIFT;
      pages = 1 << (8 - DG_PSIZE_10_SHIFT);
      banks = 1 << DG_PSIZE_10_SHIFT;
      lin_lut_size = sizeof (lin_lut->lut10data);
      break;
    case 8:
      shift = DG_PSIZE_8_SHIFT;
      pages = 1 << (8 - DG_PSIZE_8_SHIFT);
      banks = 1 << DG_PSIZE_8_SHIFT;
      lin_lut_size = sizeof (lin_lut->lut8data);
      break;
    }
  
  if ((lin_lut = iti_malloc(lin_lut_size, ITI_PAGED)) == NULL)
    {
      iti->errno = -EITI_POINTER;
      goto error;
    }
  ITI_PRINT3A("    lin_lut: 0x%p\n", lin_lut);

  for (count = 0; count < banks * LUT16_BANK_SIZE; count++)
    {
      switch (mode)
	{
	case ITI_MODE_LINEAR:
	  data = count;
	  break;
	case ITI_MODE_NORM_TO8:
	  data = count >> shift;
	  break;
	case ITI_MODE_NORM_TO16:
	  data = count << (8 - shift);
	  break;
	}
      /* this is a union, so we assign the biggest member */
      lin_lut->lut16data[count] = data;      
    }
  
  SET_LUTPG_BY_DEVICE (0);
  if (iti_set_lut16 (iti, device, lin_lut) < 0) goto error;

  iti_memset (lin_lut, 0x00, lin_lut_size);

  for (page = 1; page < pages; page++)
    {
      SET_LUTPG_BY_DEVICE (page << shift);
      if (iti_set_lut16 (iti, device, lin_lut) < 0) goto error;
    }
  
  SET_LUTPG_BY_DEVICE (0);
  ITI_PRINT3A("    lin_lut: 0x%p\n", lin_lut);
  iti_free (lin_lut, ITI_PAGED);
  return OK;

error:
  ITI_PRINT3A("    lin_lut: 0x%p\n", lin_lut);
  iti_free (lin_lut, ITI_PAGED);
  return ERROR;  
}

/*
 * icp_amvs_startup
 */

int
icp_amvs_startup (itifg_t *iti)
{
  ITI_PRINT1("> icp_amvs_startup\n");

  iti->errno = EITI_ERR_NO;

  /* initialization */
  ITI_PRINT1A("    AM-VS revision: %d\n", vs_get_rev (iti));
      
  /* preset reserved bits to a defined (zero) state */
  vs_init_con1 (iti);
  vs_init_con2 (iti);
  vs_init_trigcon (iti);
  vs_init_cpcon (iti);
  
  /* ADC needs special handling */
  vs_init_adccon (iti);
  
  /* initialize lookup table */
  vs_lin_lut (iti);
  
  /* check for errors */
  if (iti->errno < 0)
    {
      iti_error (iti, "icp_amvs_startup");
      return -EIO;
    }
  
  /* set pixelsize */
  if (cr_set_pixsz (iti, CR_PIXSZ_8) < 0)
    {
      iti_error (iti, "cr_set_pixsz");
      return -EIO;
    }

  ITI_PRINT1("< icp_amvs_startup\n");  
  return OK;
}

/*
 * icp_amvs_init
 */

int
icp_amvs_init (itifg_t *iti)
{
  struct vs_cam_t *camcnf;

  ITI_PRINT1("> icp_amvs_init\n");

  iti->errno = EITI_ERR_NO;

  camcnf = &iti->dev.vs_cam;

  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("1\n");
  vs_set_tsel (iti, camcnf->tsel);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("2\n");
  vs_set_clksrc (iti, camcnf->clksrc);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("3\n");
  vs_set_lpfsel (iti, camcnf->lpfsel);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("4\n");
  vs_set_iscale (iti, camcnf->iscale);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("5\n");
  vs_set_smode (iti, camcnf->smode);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("6\n");

  vs_set_tsrc (iti, camcnf->tsrc);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("7\n");
  vs_set_lenpol (iti, camcnf->lenpol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("8\n");
  vs_set_fldpol (iti, camcnf->fldpol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("9\n");
  vs_set_fenpol (iti, camcnf->fenpol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("10\n");
  vs_set_pclkpol (iti, camcnf->pclkpol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("11\n");
  vs_set_eclken (iti, camcnf->eclken);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("12\n");
  vs_set_fldsten (iti, camcnf->fldsten);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("13\n");
  vs_set_vsten (iti, camcnf->vsten);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("14\n");

  vs_set_trigen (iti, camcnf->trgen);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("15\n");
  vs_set_trigsrc (iti, camcnf->trgsrc);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("16\n");
  vs_set_fmode (iti, camcnf->fmode);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("17\n");
  vs_set_tmode (iti, camcnf->tmode);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("18\n");
  vs_set_strbdly (iti, camcnf->strbdly);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("19\n");

  vs_set_hoff (iti, camcnf->hoff);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("20\n");
  vs_set_hsz (iti, camcnf->hsz);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("21\n");
  vs_set_voff (iti, camcnf->voff);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("22\n");
  vs_set_vsz (iti, camcnf->vsz);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("23\n");
  vs_set_froff (iti, camcnf->froff);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("24\n");

  vs_set_cppos (iti, camcnf->cppos);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("25\n");
  vs_set_cpen (iti, camcnf->cpen);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("26\n");
  vs_set_cpsrc (iti, camcnf->cpsrc);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("27\n");
  vs_set_strbsel (iti, camcnf->strbsel);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("28\n");
  vs_set_strbpol (iti, camcnf->strbpol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("29\n");
  vs_set_cpst (iti, camcnf->cpst);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("30\n");
  
  vs_set_nref (iti, camcnf->nref);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("31\n");
  vs_set_pref (iti, camcnf->pref);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("32\n");

#if (DEBUG_ITIFG & 4)
  vs_dump_regs (iti);
#endif

  /* check for errors and restore default structure */
  if (iti->errno != EITI_ERR_NO)
    {
      iti_memcpy (&iti->dev.vs_cam, &vs_def_cam,
		  sizeof(struct vs_cam_t));
      iti_error (iti, "icp_amvs_init");
      return -EIO;
    }

  /* set operation mode (interlaced or not) */
  if (vs_get_smode (iti) == DG_ILACE)
    {
      if (cr_set_fldsel (iti, CR_EVEN_FIELD) < 0)
	{
	  iti_error (iti, "cr_set_fldsel");
	  return -EIO;
	}
    }
  else
    {
    if (cr_set_fldsel (iti, CR_NEXT_FIELD) < 0)
      {
	iti_error (iti, "cr_set_fldsel");
	return -EIO;
      }
    }

  ITI_PRINT1("< icp_amvs_init\n");
  return OK;
}

/*
 * icp_amdg_startup
 */

int
icp_amdg_startup (itifg_t *iti)
{
#if (DEBUG_ITIFG & 2)
  u_char version_number = 15;
  
  const char *version[16] =
  {
    "16bit-RS422",
    "16bit-TTL",
    "24bit-TTL",
    "8bit-RS422",
    "8bit-TTL",
  };
  
  version_number = dg_get_modver (iti);
#endif
    
  ITI_PRINT1("> icp_amdg_startup\n");

  iti->errno = EITI_ERR_NO;

  /* initialization */
  ITI_PRINT1A("    AM-DIG version: %s\n", version_number > 4 ?
	      "    unknown version" : version[version_number]);
  ITI_PRINT1A("          revision: %d\n", dg_get_rev (iti));
  
  /* preset reserved bits to a defined (zero) state */
  dg_init_clkcon (iti);
  dg_init_ctcon (iti);
  dg_init_excon (iti);
  dg_init_lutcon (iti);
  dg_init_lutsel (iti);
  dg_init_trigcon (iti);
  
  /* check for errors */
  if (iti->errno < 0)
    {
      iti_error (iti, "icp_amdg_startup");
      return -EIO;
    }
  
  /* set pixelsize */
  switch (dg_get_modver (iti))
    {
    case DG_MODVER_8_TTL:
    case DG_MODVER_8_RS422:
      if (cr_set_pixsz (iti, CR_PIXSZ_8) < 0)
	{
	  iti_error (iti, "cr_set_pixsz");
	  return -EIO;
	}
      break;
    case DG_MODVER_16_TTL:
    case DG_MODVER_16_RS422:
      if (dg_set_psize (iti, DG_PSIZE_16) < 0)
	{
	  iti_error (iti, "dg_set_psize");
	  return -EIO;
	}
      dg_def_cam.psize = DG_PSIZE_16;
      if (dg_set_muxb (iti, DG_MUXB_HI) < 0)
	{
	  iti_error (iti, "dg_set_muxb");
	  return -EIO;
	}
      dg_def_cam.muxb = DG_MUXB_HI;
      if (cr_set_pixsz (iti, CR_PIXSZ_16) < 0)
	{
	  iti_error (iti, "cr_set_pixsz");
	  return -EIO;
	}
      break;
    case DG_MODVER_24_TTL:
      if (dg_set_psize (iti, DG_PSIZE_16) < 0)
	{
	  iti_error (iti, "dg_set_psize");
	  return -EIO;
	}
      dg_def_cam.psize = DG_PSIZE_16;
      if (dg_set_muxb (iti, DG_MUXB_HI) < 0)
	{
	  iti_error (iti, "dg_set_muxb");
	  return -EIO;
	}
      dg_def_cam.muxb = DG_MUXB_HI;
      if (IS_REV_B)
	{
	  if (cr_set_pixsz (iti, CR_PIXSZ_24_24) < 0)
	    {
	      iti_error (iti, "cr_set_pixsz");
	      return -EIO;
	    }
	}
      else
	{
	  if (cr_set_pixsz (iti, CR_PIXSZ_24_32) < 0)
	    {
	      iti_error (iti, "cr_set_pixsz");
	      return -EIO;
	    }
	}
      break;
    default:
      return -EIO;
    }
  
  /* optional features (LUT's) */
  if (dg_get_modver (iti) != DG_MODVER_8_TTL &&
      dg_get_modver (iti) != DG_MODVER_8_RS422)
    if (dg_lin_lut16 (iti, '0', ITI_MODE_LINEAR) < 0)
      {
	iti_error (iti, "dg_lin_lut16");
	return -EIO;
      }
  
  if (dg_get_modver (iti) == DG_MODVER_24_TTL)
    if (dg_lin_lut8 (iti) < 0)
      {
	iti_error (iti, "dg_lin_lut8");
	return -EIO;
      }

  ITI_PRINT1("< icp_amdg_startup\n");  
  return OK;
}

/*
 * icp_amdg_init
 */

int
icp_amdg_init (itifg_t *iti)
{
  struct dg_cam_t *camcnf;
  
  ITI_PRINT1("> icp_amdg_init\n");
  
  iti->errno = EITI_ERR_NO;
  
  camcnf = &iti->dev.dg_cam;

  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("1\n");
  dg_set_fsel (iti, camcnf->fsel);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("2\n");
  dg_set_clkdiv (iti, camcnf->clkdiv);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("3\n");
  dg_set_xillen (iti, camcnf->xillen);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("4\n");
  dg_set_mc (iti, camcnf->mc);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("5\n");

  dg_set_pclkpol (iti, camcnf->pclkpol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("6\n");
  dg_set_lenpol (iti, camcnf->lenpol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("7\n");
  dg_set_fenpol (iti, camcnf->fenpol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("8\n");
  dg_set_fldpol (iti, camcnf->fldpol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("9\n");
  dg_set_smode (iti, camcnf->smode);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("10\n");
  dg_set_ilmode (iti, camcnf->ilmode);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("11\n");
  dg_set_lensm (iti, camcnf->lensm);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("12\n");
  dg_set_nolmiss (iti, camcnf->nolmiss);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("13\n");
  dg_set_lvar (iti, camcnf->lvar);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("14\n");

  dg_set_exclk (iti, camcnf->exclk);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("15\n");
  dg_set_extmd (iti, camcnf->extmd);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("16\n");
  dg_set_exsen (iti, camcnf->exsen);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("17\n");
  dg_set_exspol (iti, camcnf->expol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("18\n");
  dg_set_prien (iti, camcnf->prien);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("19\n");
  dg_set_prip (iti, camcnf->prip);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("20\n");
  dg_set_extime (iti, camcnf->extime);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("21\n");

  dg_set_psize (iti, camcnf->psize);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("22\n");
  dg_set_ilpixel (iti, camcnf->ilpixel);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("23\n");
  dg_set_muxa (iti, camcnf->muxa);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("24\n");
  dg_set_muxb (iti, camcnf->muxb);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("25\n");
  dg_set_muxc (iti, camcnf->muxc);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("26\n");

  dg_set_hoff (iti, camcnf->hoff);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("27\n");
  dg_set_hact (iti, camcnf->hact);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("28\n");
  dg_set_voff (iti, camcnf->voff);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("29\n");
  dg_set_vact (iti, camcnf->vact);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("30\n");

  dg_set_trigen (iti, camcnf->trigen);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("31\n");
  dg_set_trigsrc (iti, camcnf->trigsrc);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("32\n");

#if (DEBUG_ITIFG & 4)
  dg_dump_regs (iti);
#endif

  /* check for errors and restore default structure */
  if (iti->errno != EITI_ERR_NO)
    {
      iti_memcpy (&iti->dev.dg_cam, &dg_def_cam,
		  sizeof(struct dg_cam_t));
      iti_error (iti, "icp_amdg_init");
      return -EIO;
    }

  /* set operation mode (interlaced or not) */
  if (dg_get_ilmode (iti) == DG_ILACE)
    {
      if (cr_set_fldsel (iti, CR_EVEN_FIELD) < 0)
	{
	  iti_error (iti, "cr_set_fldsel");
	  return -EIO;
	}
    }
  else
    {
    if (cr_set_fldsel (iti, CR_NEXT_FIELD) < 0)
      {
	iti_error (iti, "cr_set_fldsel");
	return -EIO;
      }
    }

  ITI_PRINT1("< icp_amdg_init\n");
  return OK;
}

/*
 * icp_ampv_startup
 */

int
icp_ampv_startup (itifg_t *iti)
{
  ITI_PRINT1("> icp_ampv_startup\n");

  iti->errno = EITI_ERR_NO;
  
  /* initialization */
  ITI_PRINT1("    PCVision revision\n");
  ITI_PRINT1A("      `motherboard': %d\n",
	      iti->revision - PV_REV_MIN);
  ITI_PRINT1A("    `daughterboard': %d\n", pv_get_rev (iti));
  
  /* preset reserved bits to a defined (zero) state */
  pv_init_vsyncpw (iti);
  pv_init_strbdly (iti);
  pv_init_frstdly (iti);
  pv_init_con1 (iti);
  pv_init_con2 (iti);
  pv_init_trigcon (iti);
  pv_init_p_clamp (iti);
  
  /* ADC needs special handling */
  pv_init_adccon (iti);
  
  /* init pll/xtal clock generation circuit */
  pv_init_pllregs (iti);
  
  /* initialize lookup table */
  pv_lin_lut (iti);
  
  /* check for errors */
  if (iti->errno < 0)
    {
      iti_error (iti, "icp_ampv_startup");
      return -EIO;
    }
  
  /* set pixelsize */
  if (cr_set_pixsz (iti, CR_PIXSZ_8) < 0)
    {
      iti_error (iti, "cr_set_pixsz");
      return -EIO;
    }

  ITI_PRINT1("< icp_ampv_startup\n");      
  return OK;
}

/*
 * icp_ampv_init
 */

int
icp_ampv_init (itifg_t *iti)
{
  struct pv_cam_t *camcnf;
  
  ITI_PRINT1("> icp_ampv_init\n");
  
  iti->errno = EITI_ERR_NO;
  
  camcnf = &iti->dev.pv_cam;

  pv_set_htotal (iti, camcnf->htotal);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("2-0\n");
  pv_set_hsyncpw (iti, camcnf->hsyncpw);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("2-1\n");
  pv_set_hequipw (iti, camcnf->hequipw);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("2-2\n");
  pv_set_hserrpw (iti, camcnf->hserrpw);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("2-3\n");

  pv_set_vtotal (iti, camcnf->vtotal);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("3-0\n");
  pv_set_vsyncpw (iti, camcnf->vsyncpw);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("3-1/A\n");
  pv_set_hsyncpol (iti, camcnf->hsyncpol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("3-1/B\n");
  pv_set_vsyncpol (iti, camcnf->vsyncpol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("3-1/C\n");
  pv_set_vgstart (iti, camcnf->vgstart);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("3-2\n");
  pv_set_vgwidth (iti, camcnf->vgwidth);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("3-3\n");
  pv_set_hoffset (iti, camcnf->hoffset);

  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("4-0\n");
  pv_set_hactive (iti, camcnf->hactive);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("4-1\n");
  pv_set_voffset (iti, camcnf->voffset);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("4-2\n");
  pv_set_vactive (iti, camcnf->vactive);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("4-3\n");

  
  pv_set_strbdly (iti, camcnf->strbdly);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("5-0/A\n");
  if (iti->revision > PV_REV_A)
    {
      pv_set_lenpol (iti, camcnf->lenpol);
      if (iti->errno == EITI_ERR_NO) ITI_PRINT2("5-0/B\n");
      pv_set_fenpol (iti, camcnf->fenpol);
      if (iti->errno == EITI_ERR_NO) ITI_PRINT2("5-0/C\n");
    }
  pv_set_froff (iti, camcnf->froff);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("5-1/A\n");
  pv_set_frstonv (iti, camcnf->frstonv);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("5-1/B\n");
  pv_set_frstmd (iti, camcnf->frstmd);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("5-1/C\n");
  pv_set_frstpol (iti, camcnf->frstpol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("5-1/D\n");
  pv_set_frstsz (iti, camcnf->frstsz);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("5-1/E\n");
  pv_set_syncsel (iti, camcnf->syncsel);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("5-2/A\n");
  pv_set_chsyncen (iti, camcnf->chsyncen);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("5-2/B\n");
  pv_set_vsyncen (iti, camcnf->vsyncen);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("5-2/C\n");
  pv_set_clksrc (iti, camcnf->clksrc);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("5-2/D\n");
  pv_set_lpfsel (iti, camcnf->lpfsel);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("5-2/E\n");
  if (iti->revision > PV_REV_A)
    {
      pv_set_vscan (iti, camcnf->vscan);
      if (iti->errno == EITI_ERR_NO) ITI_PRINT2("5-2/F\n");
    }
  pv_set_iscale (iti, camcnf->iscale);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("5-2/G\n");
  if (iti->revision > PV_REV_A) {
    pv_set_vclkpol (iti, camcnf->vclkpol);
    if (iti->errno == EITI_ERR_NO) ITI_PRINT2("5-2/H\n");
  }
  pv_set_smode (iti, camcnf->smode);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("5-2/I\n");
  pv_set_camsel (iti, camcnf->camsel);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("5-2/J\n");
  pv_set_fldshft (iti, camcnf->fldshft);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("5-2/K\n");

  pv_set_trigen (iti, camcnf->trigen);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("6-0/A\n");
  pv_set_trigsrc (iti, camcnf->trigsrc);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("6-0/B\n");
  pv_set_trigpol (iti, camcnf->trigpol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("6-0/C\n");
  pv_set_strbmd (iti, camcnf->strbmd);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("6-0/D\n");
  pv_set_strbpol (iti, camcnf->strbpol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("6-0/E\n");
  if (iti->revision > PV_REV_A)
    {
      pv_set_strbpol (iti, camcnf->strbpol);
      if (iti->errno == EITI_ERR_NO) ITI_PRINT2("6-0/F\n");
    }
  pv_set_skpfldmd (iti, camcnf->skpfldmd);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("6-0/G\n");
  pv_set_ccnt (iti, camcnf->ccnt);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("6-2/A\n");
  pv_set_pclmpol (iti, camcnf->pclmpol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("6-2/B\n");
  pv_set_clmpsrc (iti, camcnf->clmpsrc);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("6-2/C\n");

  pv_set_fdiv (iti, camcnf->fdiv);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("8-0/A\n");
  pv_set_rdiv (iti, camcnf->rdiv);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("8-3/A\n");
  pv_set_vcogain (iti, camcnf->vcogain);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("8-4/A\n");
  pv_set_pda (iti, camcnf->pda);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("8-5/E\n");
  pv_set_lcnt (iti, camcnf->lcnt);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("8-6/A\n");

  pv_set_nref (iti, camcnf->nref);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("9-0\n");
  pv_set_pref (iti, camcnf->pref);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("9-1\n");

#if (DEBUG_ITIFG & 4)
  pv_dump_regs (iti);
#endif

  /* check for errors and restore default structure */
  if (iti->errno != EITI_ERR_NO)
    {
      iti_memcpy (&iti->dev.pv_cam, &pv_def_cam,
		  sizeof(struct pv_cam_t));
      iti_error (iti, "icp_ampv_init");
      return -EIO;
    }

  /* set operation mode (interlaced or not) */
  if (pv_get_smode (iti) == PV_ILACE)
    {
      if (cr_set_fldsel (iti, CR_EVEN_FIELD) < 0)
	{
	  iti_error (iti, "cr_set_fldsel");
	  return -EIO;
	}
    } 
  else 
    {
      if (cr_set_fldsel (iti, CR_NEXT_FIELD) < 0)
	{
	  iti_error (iti, "cr_set_fldsel");
	  return -EIO;
	}
    }

  ITI_PRINT1("> icp_ampv_init\n");  
  return OK;
}

/*
 * pcd_viam_startup
 */

int
pcd_viam_startup (itifg_t *iti)
{
  int board, found = FALSE;

  u_char pcd_brdcfg;

#if (DEBUG_ITIFG & 2)
  u_char sform_number;
  u_char pclk_polarity;
  
  const char *sform_string[8] =
  {
    "8bit single tap",
    "10,12,14,16bit single tap",
    "24bit packed single tap",
    "32bit single tap",
    "8bit dual tap",
    "8bit quad tap",
    "10,12,14,16bit dual tap",
    "reserved or error"
  };
#endif
  
  ITI_PRINT1("> pcd_viam_startup\n");

  pcd_brdcfg = pcd_get_brdid (iti);
  /* initialization */
  switch (pcd_brdcfg)
    {
    case PCD_BRDCFG_4MB_RS422:
      ITI_PRINT1("    PC-DIG configuration: 4MB/RS422\n");
      break;
    case PCD_BRDCFG_4MB_LVDS:
      ITI_PRINT1("    PC-DIG configuration: 4MB/LVDS\n");
      break;
    default:
      iti_printna ("    PC-DIG configuration: not known (%d)\n", pcd_brdcfg);
    }

  /* stop here, if not initialized */
  if (!FPGA_OK)
    return OK;

  iti->errno = EITI_ERR_NO;

#if (DEBUG_ITIFG & 2)
  sform_number = pcd_get_sform (iti);
  if (sform_number > 7)
    sform_number = 7;
  pclk_polarity = pcd_get_pclkpol (iti);
#endif
    
  ITI_PRINT1("    PC-DIG storage format:\n");
  ITI_PRINT1A("      %s\n", sform_string[sform_number]);
  ITI_PRINT1("    PC-DIG clock polarity:\n");
  ITI_PRINT1A("      %s\n", pclk_polarity ? "Rising Edge" : "Falling Edge");
  
  /* preset reserved bits to a defined (zero) state */
  pcd_init_ctcon (iti);
  pcd_init_outcon (iti);
  pcd_init_exscon (iti);
  pcd_init_cclkcon (iti);
  pcd_init_lutcon0 (iti);
  pcd_init_lutcon1 (iti);
  pcd_init_trigcon (iti);
  pcd_init_trigdiv (iti);

  /* the first board is master and drives all other boards */
  board = 0;
  while (itifg[board] != NULL && itifg[board] != iti)
    if (itifg[board++]->device_id == ITI_PCDIG)
      found = TRUE;

  if (found)
    {
      ITI_PRINT1("    This board is a slave\n");
      pcd_set_driveext (iti, PCD_DISABLE);
      pcd_set_useext (iti, PCD_ENABLE);
    }
  else
    {
      ITI_PRINT1("    This board is the master\n");
      pcd_set_driveext (iti, PCD_ENABLE);
      pcd_set_useext (iti, PCD_DISABLE);
    }
  
  /* check for errors */
  if (iti->errno < 0)
    {
      iti_error (iti, "pcd_viam_startup");
      return -EIO;
    }
  
  /* linearize lookup tables */
  if (pcd_lin_lut (iti, '0', ITI_MODE_LINEAR))
    {
      iti_error (iti, "pcd_lin_lut0");
      return -EIO;
    }
  if (pcd_lin_lut (iti, '1', ITI_MODE_LINEAR))
    {
      iti_error (iti, "pcd_lin_lut1");
      return -EIO;
    }

  ITI_PRINT1("< pcd_viam_startup\n");
  return OK;
}

/*
 * pcd_viam_init
 */

int
pcd_viam_init (itifg_t *iti)
{
  int board, pcdig;
  struct pcd_cam_t *camcnf;

  u_long clkvalue;
  u_char exsclk = 0;
  u_long exstime = 0, midpnt = 0;
  double exstime_convert = 0.0, midpnt_convert = 0.0;
  
  ITI_PRINT1("> pcd_viam_init\n");
  
  /* stop here, if not initialized */
  if (!FPGA_OK)
    return OK;

  iti->errno = EITI_ERR_NO;
  
  camcnf = &iti->dev.pcd_cam;

  pcd_set_lenenb (iti, camcnf->lenenb);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("1/A\n");
  pcd_set_lensync (iti, camcnf->lensync);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("1/B\n");
  pcd_set_lenpol (iti, camcnf->lenpol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("1/C\n");
  pcd_set_fenenb (iti, camcnf->fenenb);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("1/D\n");
  pcd_set_fenpol (iti, camcnf->fenpol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("1/E\n");
  pcd_set_multenb (iti, camcnf->multenb);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("1/F\n");
  pcd_set_multpol (iti, camcnf->multpol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("1/G\n");
  pcd_set_multmd (iti, camcnf->multmd);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("1/H\n");
  pcd_set_restime (iti, camcnf->restime);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("2/A\n");
  pcd_set_respol (iti, camcnf->respol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("2/B\n");
  pcd_set_cam3enb (iti, camcnf->cam3enb);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("2/C\n");
  pcd_set_exsenb (iti, camcnf->exsenb);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("2/D\n");
  pcd_set_exspol (iti, camcnf->exspol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("2/E\n");
  pcd_set_exsmd (iti, camcnf->exsmd);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("2/F\n");
  pcd_set_prienb (iti, camcnf->prienb);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("2/G\n");
  pcd_set_pripol (iti, camcnf->pripol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("2/H\n");
  pcd_set_cammd (iti, camcnf->cammd);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("2/I\n");
  pcd_set_cclkenb (iti, camcnf->cclkenb);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("2/J\n");
  pcd_set_cclkpol (iti, camcnf->cclkpol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("2/K\n");
  pcd_set_exssel (iti, camcnf->exssel);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("3/A\n");
  pcd_set_prisel (iti, camcnf->prisel);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("3/B\n");
  pcd_set_prihold (iti, camcnf->prihold);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("3/C\n");
  pcd_set_hoff (iti, camcnf->hoff);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("4\n");
  pcd_set_hact (iti, camcnf->hact);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("5\n");
  pcd_set_voff (iti, camcnf->voff);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("6\n");
  pcd_set_vact (iti, camcnf->vact);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("7\n");
  pcd_set_lut0psz (iti, camcnf->lut0psz);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("8/A\n");
  pcd_set_lut0pg (iti, camcnf->lut0pg);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("8/B\n");
  pcd_set_lut1psz (iti, camcnf->lut1psz);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("9/A\n");
  pcd_set_lut1pg (iti, camcnf->lut1pg);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("9/B\n");
  pcd_set_trig0pol (iti, camcnf->trig0pol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("10/A\n");
  pcd_set_trig0src (iti, camcnf->trig0src);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("10/B\n");
  pcd_set_trig0gd (iti, camcnf->trig0gd);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("10/C\n");
  pcd_set_trig1pol (iti, camcnf->trig1pol);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("10/D\n");
  pcd_set_trig1src (iti, camcnf->trig1src);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("10/E\n");
  pcd_set_trig1gd (iti, camcnf->trig1gd);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("10/F\n");
  pcd_set_trigdiv (iti, camcnf->trigdiv);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("11/A\n");
  pcd_set_divcon (iti, camcnf->divcon);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("11/B\n");

  /* set the exposure/readout (exsclk/exstime/midpnt) cycle times */
  exstime_convert = camcnf->extime * 1000 + camcnf->rotime * 1000;
  midpnt_convert = camcnf->rotime * 1000;
  exstime_convert /= ITI_DLY_BASE, midpnt_convert /= ITI_DLY_BASE;
  exstime = exstime_convert, midpnt = midpnt_convert;
  if (exstime > (1 << 10) * 65537)
    exsclk = PCD_EXSCLK_3_496_k, exstime >>= 12, midpnt >>= 12;
  else
    if (exstime > (1 << 8) * 65537)
      exsclk = PCD_EXSCLK_13_982_k, exstime >>= 10, midpnt >>= 10; 
    else
      if (exstime > (1 << 6) * 65537)
	exsclk = PCD_EXSCLK_55_930_k, exstime >>= 8, midpnt >>= 8; 
      else
	if (exstime > (1 << 4) * 65537)
	  exsclk = PCD_EXSCLK_223_719_k, exstime >>= 6, midpnt >>= 6; 
	else
	  if (exstime > (1 << 2) * 65537)
	    exsclk = PCD_EXSCLK_894_875_k, exstime >>= 4, midpnt >>= 4; 
	  else
	    if (exstime > (1 << 1) * 65537)
	      exsclk = PCD_EXSCLK_3_580_M, exstime >>= 2, midpnt >>= 2; 
	    else
	      if (exstime > 65537)
		exsclk = PCD_EXSCLK_7_159_M, exstime >>= 1, midpnt >>= 1; 
  /* update the exstime software and hardware without interrupts! */
  /* find first board */
  board = 0;
  while (itifg[board] != NULL && itifg[board] != iti)
    board++;
  for (pcdig = board; pcdig >= 0; pcdig--)
    if (itifg[pcdig]->device_id == ITI_PCDIG)
      if (pcd_get_driveext (itifg[pcdig]) == PCD_ENABLE)
	break;
  if (itifg[pcdig] < 0)
    return -EINVAL;
  /* do it at the first pcdig */
  iti_irq_lock (itifg[pcdig]);
  new_exsclk = exsclk;
  new_exstime = exstime;
  new_midpnt = midpnt;
  pcd_set_exsenb (itifg[pcdig], PCD_DISABLE);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("12/A\n");
  pcd_set_exsclk (itifg[pcdig], exsclk);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("12/B\n");
  pcd_set_exstime (itifg[pcdig], exstime);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("12/C\n");
  pcd_set_emidpnt (itifg[pcdig], (midpnt / (1 << PCD_EMIDPNT_SHIFT)) *
		   (1 << PCD_EMIDPNT_SHIFT));
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("12/D\n");
  pcd_set_pmidpnt (itifg[pcdig], (midpnt / (1 << PCD_PMIDPNT_SHIFT)) *
		   (1 << PCD_PMIDPNT_SHIFT));
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("12/E\n");
  pcd_set_exsenb (itifg[pcdig], PCD_ENABLE);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("12/F\n");
  /* the values are updated, no need to set them again */
  need_update = FALSE;
  iti_irq_unlock (itifg[pcdig]);

  /* set the camera clock */
  pcd_calc_clkval (iti, camcnf->cclkfreq, &clkvalue);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("13/A\n");
  pcd_write_clkval (iti, 'A', clkvalue);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("13/B\n");
  pcd_set_synmuxa (iti, PCD_SYNMUX_PROG);
  if (iti->errno == EITI_ERR_NO) ITI_PRINT2("13/C\n");

#if (DEBUG_ITIFG & 4)
  pcd_dump_regs (iti);
#endif
  
  /* check for errors and restore default structure */
  if (iti->errno != EITI_ERR_NO)
    {
      iti_memcpy (&iti->dev.pcd_cam, &pcd_def_cam,
		  sizeof(struct pcd_cam_t));
      iti_error (iti, "pcd_viam_init");
      return -EIO;
    }

  ITI_PRINT1("> pcd_viam_init\n");
  return OK;
}

