/*		 ITI-FG  Kernel module free public release
 *
 *		     Copyright (C) 1996-2000 GOM mbH.
 *	         Rebenring 33, 38106 Braunschweig, GERMANY
 *		            All rights reserved
 *
 * $Id: osLinux2X.c,v 7.0.2.1 2000/11/29 15:02:59 cvs Exp $
 *
 */

/*
 * Includes:
 */

#include "itifgOS.h"	/* os specific stuff */
#include "itifgExt.h"	/* offical driver interface */
#include "itifgReg.h"	/* register definitions ICR Regs */
#include "icpciReg.h"	/* register definitions IC-PCI */
#include "amvsReg.h"	/* register definitions AM-VS */
#include "amdigReg.h"	/* register definitions AM-DIG */
#include "ampcvReg.h"	/* register definitions PCVision */
#include "pcdigReg.h"	/* register definitions PC-DIG */
#include "itifgIntD.h"	/* driver internals */
#include "itifgIntF.h"	/* internal functions */

#include <linux/module.h>	/* init_/cleanup_ module */

#ifdef MODULE
/* 
 * Command Line Options (insmod)
 *
 * Major Device Number assignement:
 * 'major0=x',  0  (2.0|2.2): dynamic assignement (default)
 * 'major1=x',  1 - 128|256 : explicitely define the number (1 - MAX_CHRDEV)
 *                > 128|256 : old style behavior (ITI_MAJOR)
 *		
 */

static u_int major = 0;

#if LINUX_VERSION_CODE >= 0x020200
MODULE_PARM(major, "1-" __MODULE_STRING(MAX_CHRDEV) "i");
MODULE_DESCRIPTION("Imaging Technology Inc. frame grabber device driver");
MODULE_AUTHOR("Matthias Stein <M.Stein@gom.com>");
#endif

#define iti_os_detach	cleanup_module 
#define iti_os_attach	init_module 
#endif

/*
 * Defines
 */

#if LINUX_VERSION_CODE < 0x020100
#define ITI_LATENCY_DEF	32	/* default pci bus master latency */
#endif

/*
 * Globals:
 */

extern int iti_version_code;	/* kept in itifgAux.c for consistency */
extern const struct mod_names_t mod_names[4]; /* also */

pgprot_t protection_map[16] = {
	__P000, __P001, __P010, __P011, __P100, __P101, __P110, __P111,
	__S000, __S001, __S010, __S011, __S100, __S101, __S110, __S111
};

/*
 * Prototypes:
 */

static void
iti_os_interrupt (int, void*, struct pt_regs*);
static void
iti_os_cam_timeout (u_long);
static void
iti_os_iop_timeout (u_long);
static void
iti_os_task_queue (void*);

#if LINUX_VERSION_CODE < 0x020100
static int
iti_os_open (struct inode*, struct file*);
static void
iti_os_close (struct inode*, struct file*);
static int
iti_os_read (struct inode*, struct file*, char*, int);
static int
iti_os_write (struct inode*, struct file*, const char*, int);
static int
iti_os_lseek (struct inode*, struct file*, off_t, int);
static int
iti_os_ioctl (struct inode*, struct file*, u_int, u_long); 
static int
iti_os_mmap (struct inode*, struct file*, struct vm_area_struct*);
static int
iti_os_select (struct inode*, struct file*, int, select_table*);
static int
iti_os_fasync (struct inode*, struct file*, int);
#elif LINUX_VERSION_CODE >= 0x020200
static int
iti_os_open (struct inode*, struct file*);
static int
iti_os_release (struct inode*, struct file*);
static ssize_t
iti_os_read (struct file*, char*, size_t, loff_t*);
static ssize_t
iti_os_write (struct file*, const char*, size_t, loff_t*);
static loff_t
iti_os_llseek (struct file*, loff_t, int);
static int
iti_os_ioctl (struct inode*, struct file*, unsigned int, unsigned long); 
static int
iti_os_mmap (struct file*, struct vm_area_struct*);
static u_int
iti_os_poll (struct file*, struct poll_table_struct*);
static int
iti_os_fasync (int, struct file*, int);
#endif

static struct file_operations iti_fops =
{
#if LINUX_VERSION_CODE < 0x020100
  lseek:	iti_os_lseek,
  read:		iti_os_read,
  write:	iti_os_write,
  NULL,		/* readdir */
  select:	iti_os_select,
  ioctl:	iti_os_ioctl,
  mmap:		iti_os_mmap,
  open:		iti_os_open,
  release:	iti_os_close,
  NULL,		/* fsync */
  fasync:	iti_os_fasync,
  NULL,		/* check_media_change */
  NULL		/* revalidate */
#elif LINUX_VERSION_CODE >= 0x020200
  llseek:	iti_os_llseek,
  read:		iti_os_read,
  write:	iti_os_write,
  NULL,		/* readdir */
  poll:		iti_os_poll,
  ioctl:	iti_os_ioctl,
  mmap:		iti_os_mmap,
  open:		iti_os_open,
  NULL,		/* flush */
  release:	iti_os_release,
  NULL,		/* fsync */
  fasync:	iti_os_fasync,
  NULL,		/* check_media_change */
  NULL,		/* revalidate */
  NULL		/* lock */
#endif
};

#if LINUX_VERSION_CODE < 0x020300
static int
iti_os_read_proc (char*, char**, off_t, int, int);
#elif LINUX_VERSION_CODE >= 0x020400
static int
iti_os_read_proc (char*, char**, off_t, int, int*, void*);
#endif

static int
iti_os_proc_info (char *buffer, int *length);

#if LINUX_VERSION_CODE < 0x020300
static struct proc_dir_entry iti_proc_entry =
{
  0,
  5, PFS_STRING,
  S_IFREG | S_IRUGO,
  1, 0, 0,
  0,
  NULL,
  iti_os_read_proc
};
#elif LINUX_VERSION_CODE >= 0x020400
static struct proc_dir_entry *iti_proc_entry = NULL;
#endif

/*
 * iti_malloc
 */

void* 
iti_malloc (size_t size, iti_alloc_t mode)
{
  switch (mode)
    {
    case ITI_UNPAGED:
      return kmalloc (size, GFP_KERNEL);
      break;
    case ITI_PAGED:
      return vmalloc (size);
      break;
    default:
      return NULL;
    }
  
  return NULL;
}

/*
 * iti_free
 */

void
iti_free(void *base, iti_alloc_t mode)
{
  switch (mode)
    {
    case ITI_UNPAGED:
      return kfree (base);
      break;
    case ITI_PAGED:
      return vfree (base);
      break;
    }
}

/*
 * iti_add_timer
 */

int
iti_add_timer (itifg_t *iti, time_t timeout, iti_sleep_t mode)
{
  struct timer_list *timer_ptr = NULL;
  void (*timer_func) (u_long) = NULL;

  switch (mode)
    {
    case ITI_CAMERA:
      timer_ptr = &iti->osp.cam_timer;
      timer_func = iti_os_cam_timeout;
      break;
    case ITI_IOPORT:
      timer_ptr = &iti->osp.iop_timer;
      timer_func = iti_os_iop_timeout;
      break;
    default:
      return -EINVAL;
    }

  if (!del_timer (timer_ptr))
    {
#if (0)
      ITI_PRINTR("\n");
      ITI_PRINTI("timer is not queued.\n");
#endif
    }

  timer_ptr->data = (u_long)iti;
  timer_ptr->function = timer_func;
  timer_ptr->expires = jiffies + timeout;
  add_timer (timer_ptr);

  return OK;
}

/*
 * iti_del_timer
 */

int
iti_del_timer (itifg_t *iti, iti_sleep_t mode)
{
  struct timer_list *timer_ptr = NULL;

  switch (mode)
    {
    case ITI_CAMERA:
      timer_ptr = &iti->osp.cam_timer;
      break;
    case ITI_IOPORT:
      timer_ptr = &iti->osp.iop_timer;
      break;
    default:
      return -EINVAL;
    }
  
  if (!del_timer (timer_ptr))
    {
#if (0)
      ITI_PRINTR("\n");
      ITI_PRINTI("timer is not queued.\n");
#endif
    }
  
  return OK;
}

/*
 * iti_sleep
 */

int
iti_sleep (itifg_t *iti, time_t timeout, iti_sleep_t mode)
{
  int error;

  switch (mode)
    {
    case ITI_CAMERA:
      if ((error = iti_add_timer (iti, timeout, ITI_CAMERA)))
	return error;
      if (iti->osp.cam_w_wait)
	return -EBUSY;
      iti->osp.cam_w_wait = TRUE;
      interruptible_sleep_on (&iti->osp.cam_w_queue);
      iti->osp.cam_w_wait = FALSE;
      break;
    case ITI_IOPORT:
      if ((error = iti_add_timer (iti, timeout, ITI_IOPORT)))
	return error;
      if (iti->osp.iop_w_wait)
	return -EBUSY;
      iti->osp.iop_w_wait = TRUE;
      interruptible_sleep_on (&iti->osp.iop_w_queue);
      iti->osp.iop_w_wait = FALSE;
      break;
    default:
      return -EINVAL;
    }

#if LINUX_VERSION_CODE < 0x020100
  if (current->signal & ~current->blocked)
    return -ERESTARTSYS;
#elif LINUX_VERSION_CODE >= 0x020200
  if (signal_pending(current))
    return -ERESTARTSYS;
#endif
  
  return OK;
}

/*
 * iti_block_wakeup
 */

int
iti_block_wakeup (itifg_t *iti, iti_sleep_t mode)
{
  int error;

  switch (mode)
    {
    case ITI_CAMERA:
      if (!iti->osp.cam_w_wait)
	return ERROR;
      wake_up (&iti->osp.cam_w_queue);
      if ((error = iti_del_timer (iti, ITI_CAMERA)))
	return error;
      break;
    case ITI_IOPORT:
      if (!iti->osp.iop_w_wait)
	return ERROR;
      wake_up (&iti->osp.iop_w_queue);
      if ((error = iti_del_timer (iti, ITI_IOPORT)))
	return error;
      break;
    default:
      return -EINVAL;
    }

  return OK;
}

/*
 * iti_select_wakeup
 */

int
iti_select_wakeup (itifg_t *iti, iti_sleep_t mode)
{

  switch (mode)
    {
    case ITI_CAMERA:
      if (!iti->osp.cam_s_wait)
	return ERROR;
      wake_up (&iti->osp.cam_s_queue);
      break;
    case ITI_IOPORT:
      return -ENXIO; /* can't select on ioports yet */
      break;
    default:
      return -EINVAL;
    }

  return OK;
}

/*
 * iti_marktq
 */

int
iti_marktq (itifg_t *iti)
{
  iti->osp.clr_t_queue.sync = 0;
  iti->osp.clr_t_queue.routine = iti_os_task_queue;
  iti->osp.clr_t_queue.data = (void*)iti;
  
  queue_task (&iti->osp.clr_t_queue, &tq_immediate);
  mark_bh (IMMEDIATE_BH);  

  return OK;
}

/*
 * iti_lock
 */

void
iti_lock (itifg_t *iti)
{
#if LINUX_VERSION_CODE >= 0x020200
  spin_lock (&iti->osp.spin_lock);
#endif
}

/*
 * iti_unlock
 */

void
iti_unlock (itifg_t *iti)
{
#if LINUX_VERSION_CODE >= 0x020200
  spin_unlock (&iti->osp.spin_lock);
#endif
}

/*
 * iti_irq_lock
 */

void
iti_irq_lock (itifg_t *iti)
{
#if LINUX_VERSION_CODE < 0x020100
  save_flags (iti->osp.flags);
  cli();
#elif LINUX_VERSION_CODE >= 0x020200
  spin_lock_irqsave (&iti->osp.spin_lock, iti->osp.flags);
#endif
}

/*
 * iti_irq_unlock
 */

void
iti_irq_unlock (itifg_t *iti)
{
#if LINUX_VERSION_CODE < 0x020100
  restore_flags (iti->osp.flags);
#elif LINUX_VERSION_CODE >= 0x020200
  spin_unlock_irqrestore (&iti->osp.spin_lock, iti->osp.flags);
#endif
}

/*
 * iti_get_time
 */

int
iti_get_time (struct timeval *time)
{
  if (!time)
    return ERROR;

  do_gettimeofday (time);
  return OK;
}

/*
 * iti_get_major
 */

int
iti_get_major (iti_dev_t dev, int *major)
{
  if (!major)
    return ERROR;

  *major = MAJOR(dev);
  return OK;
}

/*
 * iti_get_minor
 */

int
iti_get_minor (iti_dev_t dev, int *minor)
{
  if (!minor)
    return ERROR;

  *minor = MINOR(dev);
  return OK;
}

/*
 * iti_virt_to_phys
 */

int
iti_virt_to_phys (caddr_t arg_virt,
		  u_long *ret_phys,
		  u_long *ret_bus,
		  int user_space)
{
  pgd_t	*pgd;
  pmd_t	*pmd;
  pte_t	*pte;

  u_long virt;
  void* kvirt;

  if (user_space)
    {
      virt = (u_long)arg_virt;
      pgd = pgd_offset (current->mm, virt);
    }
  else /* kernel_space */
    {
      virt = VMALLOC_VMADDR(arg_virt);
#if LINUX_VERSION_CODE < 0x020100
      pgd = pgd_offset (current->mm, virt);
#elif LINUX_VERSION_CODE >= 0x020200
      pgd = pgd_offset_k (virt);
#endif
    }

  if (pgd_none(*pgd)) return -EFAULT;
  pmd = pmd_offset (pgd, virt);
  if (pmd_none(*pmd)) return -EFAULT;
  pte = pte_offset (pmd, virt);
  if (!pte_present(*pte)) return -EFAULT;

  kvirt = (void*)(pte_page(*pte) + (virt & ~PAGE_MASK));

  if (ret_phys)
    *ret_phys = virt_to_phys (kvirt);

  /* TESTME under 2.0 */
  if (ret_bus)
    *ret_bus = virt_to_bus (kvirt);
  
  return OK;
}

/*
 * iti_area_locked
 */

int
iti_area_locked (caddr_t  uvirt)
{
  struct vm_area_struct *vma;
      
  vma = find_vma(current->mm, (u_long)uvirt);
  if (!vma)
    return -ENOMEM;
  if (!(vma->vm_flags & VM_LOCKED))
    return -EFAULT;

  return OK;
}

/*
 * iti_lock_page
 */

int
iti_lock_page (u_long page, int user_space)
{
  void *kvirt = phys_to_virt (page);

  if (!kvirt)
    return ERROR;

#if LINUX_VERSION_CODE < 0x020100
  iti_memset (kvirt, 0, PAGE_SIZE);
#elif LINUX_VERSION_CODE >= 0x020200
#  if defined(__powerpc__) || defined(__alpha__)
  clear_page ((long)kvirt);
#  else
  clear_page (kvirt);
#  endif
#endif

#if LINUX_VERSION_CODE < 0x020300
  if (user_space)
    set_bit(PG_locked, &mem_map[MAP_NR(kvirt)].flags);
  else /* kernel space */
    set_bit(PG_reserved, &mem_map[MAP_NR(kvirt)].flags);
  atomic_inc(&(mem_map[MAP_NR(kvirt)].count));
#elif LINUX_VERSION_CODE >= 0x020400
  if (user_space)
    LockPage (virt_to_page(kvirt)); /* test (wait) and set! */
  else /* kernel space */
    SetPageReserved (virt_to_page(kvirt));
  atomic_inc (&((virt_to_page(kvirt))->count));
#endif

  return OK;
}

/*
 * iti_unlock_page
 */

int
iti_unlock_page (u_long page, int user_space)
{
  void *kvirt = phys_to_virt (page);

  if (!kvirt)
    return ERROR;

#if LINUX_VERSION_CODE < 0x020300
  if (user_space)
    clear_bit(PG_locked, &mem_map[MAP_NR(kvirt)].flags);    
  else /* kernel space */
    clear_bit(PG_reserved, &mem_map[MAP_NR(kvirt)].flags);
  atomic_dec(&(mem_map[MAP_NR(kvirt)].count));
#elif LINUX_VERSION_CODE >= 0x020400
  if (user_space)
    UnlockPage (virt_to_page(kvirt));
  else /* kernel space */
    ClearPageReserved (virt_to_page(kvirt));
  atomic_dec(&((virt_to_page (kvirt))->count));
#endif
  return OK;
}

/*
 * iti_remap_page
 */

int
iti_remap_page (caddr_t dst, u_long src)
{
  int error;
  pgprot_t pgprot;

  if (!dst || !src)
    return ERROR;
  
  pgprot_val(pgprot) = 0UL;
  pgprot = protection_map[VM_READ]; /* FIXME */
  
  if ((error = remap_page_range ((u_long)dst, src, PAGE_SIZE, pgprot)))
    return error;
  
  return (OK);
}

/*
 * iti_init_fasync
 */

int
iti_init_fasync (itifg_t *iti,
		 iti_file_t *file,
		 iti_inode_t *inode,		 
		 int desc,
		 int mode)
{
  int error;

#if LINUX_VERSION_CODE < 0x020100
  error = fasync_helper (inode, file, mode, &iti->osp.cam_a_queue);
#elif LINUX_VERSION_CODE >= 0x020200
  error = fasync_helper (desc, file, mode, &iti->osp.cam_a_queue);
#endif
  
  if (mode)
    {
      if (!iti->osp.cam_a_queue)
	return -EINVAL;
    }
  else
    if (iti->osp.cam_a_queue)
      iti->osp.cam_a_queue = NULL;
  
  return OK;
}

/*
 * iti_kill_fasync
 */

int
iti_kill_fasync (itifg_t *iti, iti_sleep_t mode)
{

  switch (mode)
    {
    case ITI_CAMERA:
      if (!iti->osp.cam_a_queue)
	return ERROR;
#if LINUX_VERSION_CODE < 0x020300
      kill_fasync (iti->osp.cam_a_queue, SIGIO);
#elif LINUX_VERSION_CODE >= 0x020400
      kill_fasync (&iti->osp.cam_a_queue, SIGIO, 0);
#endif
      break;
    case ITI_IOPORT:
      return -ENXIO;	/* not yet */
      break;
    default:
      return -EINVAL;
    }

  return OK;
}

/*
 * iti_os_interrupt
 */

void
iti_os_interrupt (int irq, void *dev_id, struct pt_regs *regs)
{
  itifg_t *iti = (itifg_t*)dev_id;

#if (0)  
  if (irq != iti->irq_line)
    iti_printn ("device irq does not match!\n");
  else
#endif
    iti_interrupt (iti);
}

/*
 * iti_os_cam_timeout
 */

void
iti_os_cam_timeout (ulong data)
{
  iti_cam_timeout ((itifg_t*)data);
}

/*
 * iti_os_iop_timeout
 */

void
iti_os_iop_timeout (ulong data)
{
  iti_iop_timeout ((itifg_t*)data);
}

/*
 * iti_os_task_queue
 */

void
iti_os_task_queue (void *data)
{
  iti_task_queue ((itifg_t*)data);
}

/*
 * iti_os_open
 */

static int
iti_os_open (struct inode *inode, struct file *file)
{
  int error;

  if ((error = iti_open (file, inode->i_rdev)))
    return error;
  
#ifdef MODULE    
  MOD_INC_USE_COUNT;
#endif    
  
  return OK;
}

/*
 * iti_os_close|release
 */

#if LINUX_VERSION_CODE < 0x020100
static void
iti_os_close (struct inode *inode, struct file *file)
{
  iti_close (file, inode, inode->i_rdev);

#ifdef MODULE    
  MOD_DEC_USE_COUNT;
#endif
}

#elif LINUX_VERSION_CODE >= 0x020200
static int
iti_os_release (struct inode *inode, struct file *file)
{
  int error;

#ifdef MODULE    
  MOD_DEC_USE_COUNT;
#endif

  if ((error = iti_close (file, inode, inode->i_rdev)))
    return error;

  return OK;
}
#endif

/*
 * iti_os_read
 */

#if LINUX_VERSION_CODE < 0x020100
static int
iti_os_read (struct inode *inode,
	     struct file *file,
	     char *buffer,
	     int count)
{
  if (verify_area(VERIFY_WRITE, buffer, count))
    return -EFAULT;

  return iti_read (file, inode->i_rdev, buffer, count);
}
#elif LINUX_VERSION_CODE >= 0x020200
static ssize_t
iti_os_read (struct file* file,
	     char *buffer,
	     size_t count,
	     loff_t *offset)
{
  return iti_read (file, file->f_dentry->d_inode->i_rdev, buffer, count);
}
#endif

/*
 * iti_os_write
 */

#if LINUX_VERSION_CODE < 0x020100
static int
iti_os_write (struct inode *inode,
	      struct file *file,
              const char *buffer,
	      int count)
{
  if (verify_area(VERIFY_READ, buffer, count))
    return -EFAULT;

  return iti_write (file, inode->i_rdev, buffer, count);
}
#elif LINUX_VERSION_CODE >= 0x020200
static ssize_t
iti_os_write (struct file *file,
	      const char *buffer,
   	      size_t count,
	      loff_t *offset)
{
  return iti_write (file, file->f_dentry->d_inode->i_rdev, buffer, count);
}
#endif

/*
 * iti_os_llseek
 */

#if LINUX_VERSION_CODE < 0x020100
static int
iti_os_lseek (struct inode *inode,
	      struct file *file,
	      off_t offset,
	      int mode)
{
  return iti_lseek (file, inode->i_rdev, offset, mode);
}
#elif LINUX_VERSION_CODE >= 0x020200
static loff_t
iti_os_llseek (struct file* file,
	       loff_t offset,
	       int mode)
{
  return iti_lseek (file, file->f_dentry->d_inode->i_rdev, offset, mode);
}
#endif

/* 
 * iti_os_ioctl
 */

#if LINUX_VERSION_CODE < 0x020100
static int
iti_os_ioctl (struct inode *inode,
	      struct file *file,
	      u_int cmd,
	      u_long arg)
#elif LINUX_VERSION_CODE >= 0x020200
static int
iti_os_ioctl (struct inode *inode,
	      struct file *file,
	      u_int cmd,
	      u_long arg)
#endif
{
  int error, count, board, funct;
  itifg_t *iti;

  union iti_inparam_t *inparam = NULL;
  union iti_outparam_t *outparam = NULL;
  
  if ((error = iti_decode_devno (inode->i_rdev, &board, &funct)))
    return error;
  iti = itifg[board];

  /* validate module specific commands */
  if (_IOC_NR(cmd) > 31 && _IOC_NR(cmd) < 224)
    switch (MODULE_ID)
      {
      case ICP_AMFA:
	if (_IOC_NR(cmd) > 63 || _IOC_TYPE(cmd) != 'G')
	  return -EINVAL;
	break;
      case ICP_AMC1:
	if (_IOC_NR(cmd) < 64 || _IOC_NR(cmd) > 95 || _IOC_TYPE(cmd) != 'G')
	  return -EINVAL;
      break;
      case ICP_AMVS:
	if (_IOC_NR(cmd) < 96 || _IOC_NR(cmd) > 127 || _IOC_TYPE(cmd) != 'G')
	  return -EINVAL;
	break;
      case ICP_AMDG:
	if (_IOC_NR(cmd) < 128 || _IOC_NR(cmd) > 159 || _IOC_TYPE(cmd) != 'G')
	  return -EINVAL;
	break;
      case ICP_AMPV:
	if (_IOC_NR(cmd) < 96 || _IOC_NR(cmd) > 127 || _IOC_TYPE(cmd) != 'H')
	  return -EINVAL;
	break;
      case ITI_VMPCD:
	if (_IOC_NR(cmd) < 128 || _IOC_NR(cmd) > 159 || _IOC_TYPE(cmd) != 'H')
	  return -EINVAL;
	break;
      default:
	return -ENXIO;
      }

  if (arg)
    {
      /* structure allocation */
      if (_IOC_DIR(cmd) == _IOC_READ ||
	  _IOC_DIR(cmd) == (_IOC_READ | _IOC_WRITE))
	{
	  outparam = iti_malloc (sizeof(union iti_outparam_t), ITI_PAGED);
	  if (outparam == NULL)
	    return -EAGAIN;
	  ITI_PRINT3A("    outparam: 0x%p\n", outparam);
	}

      if (_IOC_DIR(cmd) == _IOC_WRITE ||
	  _IOC_DIR(cmd) == (_IOC_WRITE | _IOC_READ))
	{
	  inparam = iti_malloc (sizeof(union iti_inparam_t), ITI_PAGED);
	  if (inparam == NULL)
	    return -EAGAIN;
	  ITI_PRINT3A("    inparam: 0x%p\n", inparam);
	}

      /* copyin */
      if (_IOC_DIR(cmd) == _IOC_WRITE ||
	  _IOC_DIR(cmd) == (_IOC_WRITE | _IOC_READ))
	{
	  if (cmd == GIOC_DG_SET_LUT ||
	      cmd == HIOC_PCD_SET_LUT0 || cmd == HIOC_PCD_SET_LUT1)
	    {
#if LINUX_VERSION_CODE < 0x020100
	      if (verify_area (VERIFY_READ, (void*)arg,
			       sizeof(union dg_lut16_t)))
		return -EFAULT;
#endif
	      /* data has to be copied in single-page chunks */
	      for (count = 0; count < (sizeof(union dg_lut16_t) / PAGE_SIZE);
		   count++)
		if ((error = iti_copyin ((void*)inparam + count * PAGE_SIZE, 
					 (void*)arg + count * PAGE_SIZE,
					 PAGE_SIZE)))
		  return error;
	      /* if the structure is not a multiple of PAGE_SIZE */
	      if (sizeof(union dg_lut16_t) % PAGE_SIZE)
		if ((error = iti_copyin ((void*)inparam + count * PAGE_SIZE, 
					 (void*)arg + count * PAGE_SIZE,
					 sizeof(union dg_lut16_t)%PAGE_SIZE)))
		  return error;
	    }
	  else
	    {
#if LINUX_VERSION_CODE < 0x020100
	      if (verify_area (VERIFY_READ, (void*)arg, _IOC_SIZE(cmd)))
		return -EFAULT;
#endif
	      if ((error = iti_copyin (inparam, (void*)arg, _IOC_SIZE(cmd))))
		return error;
	    }
	}
    }
  else
    if (_IOC_DIR(cmd) == _IOC_NONE)
      inparam = (void*)-1, outparam = (void*)-1;

#if LINUX_VERSION_CODE < 0x020100
  if ((error = iti_ioctl (file, inode->i_rdev, cmd, inparam, outparam)))
#elif LINUX_VERSION_CODE >= 0x020200
  if ((error = iti_ioctl (file, file->f_dentry->d_inode->i_rdev,
			  cmd, inparam, outparam)))
#endif
    {
      if (outparam != (void*)-1)
	{
	  ITI_PRINT3A("    outparam: 0x%p\n", outparam);
	  iti_free (outparam, ITI_PAGED);
	}
      if (inparam != (void*)-1)
	{
	  ITI_PRINT3A("    inparam: 0x%p\n", inparam);
	  iti_free (inparam, ITI_PAGED);
	}
      return error;
    }

  if (arg)
    {
      /* copyout */
      if (_IOC_DIR(cmd) == _IOC_READ ||
	  _IOC_DIR(cmd) == (_IOC_WRITE | _IOC_READ))
	{
	  if (cmd == GIOC_DG_GET_LUT ||
	      cmd == HIOC_PCD_GET_LUT0 || cmd == HIOC_PCD_GET_LUT1)
	    {
#if LINUX_VERSION_CODE < 0x020100
	      if (verify_area (VERIFY_WRITE, (void*)arg,
			       sizeof(union dg_lut16_t)))
		return -EFAULT;
#endif
	      /* data has to be copied in single-page chunks */
	      for (count = 0; count < (sizeof(union dg_lut16_t) / PAGE_SIZE);
		   count++)
		if ((error = iti_copyout ((void*)outparam + count * PAGE_SIZE, 
					  (void*)arg + count * PAGE_SIZE,
					  PAGE_SIZE)))
		  return error;
	      /* if the structure is not a multiple of PAGE_SIZE */
	      if (sizeof(union dg_lut16_t) % PAGE_SIZE)
		if ((error = iti_copyout ((void*)outparam + count * PAGE_SIZE, 
					  (void*)arg + count * PAGE_SIZE,
					  sizeof(union dg_lut16_t)%PAGE_SIZE)))
		  return error;
	    }
	  else
	    {
#if LINUX_VERSION_CODE < 0x020100
	      if (verify_area (VERIFY_WRITE, (void*)arg, _IOC_SIZE(cmd)))
		return -EFAULT;
#endif
	      if ((error = iti_copyout ((void*)arg, outparam, _IOC_SIZE(cmd))))
		return error;
	    }
	}

      /* free structures */
      if (outparam)
	{
	  ITI_PRINT3A("    outparam: 0x%p\n", outparam);
	  iti_free (outparam, ITI_PAGED);
	}
      if (inparam)
	{
	  ITI_PRINT3A("    inparam: 0x%p\n", inparam);
	  iti_free (inparam, ITI_PAGED);
	}
    }

  return OK;
}

/*
 * iti_os_mmap
 */

#if LINUX_VERSION_CODE < 0x020100
static int
iti_os_mmap (struct inode *inode,
	     struct file *file,
	     struct vm_area_struct *vma)
#elif LINUX_VERSION_CODE >= 0x020200
static int
iti_os_mmap (struct file *file,
	     struct vm_area_struct *vma)
#endif
{
  int error;

  caddr_t base;
  size_t size;
  
  base = (caddr_t)vma->vm_start;
  size = vma->vm_end - vma->vm_start;
  
  if (pgprot_val (vma->vm_page_prot) != pgprot_val (protection_map[VM_READ])) 
    return -EINVAL;
  
#if LINUX_VERSION_CODE < 0x020100
  if ((error = iti_mmap (file, inode->i_rdev, base, size)))
    return error;
#elif LINUX_VERSION_CODE >= 0x020200
  if ((error = iti_mmap (file, file->f_dentry->d_inode->i_rdev, base, size)))
    return error;
#endif

#if LINUX_VERSION_CODE < 0x020100
  inode->i_count++;
#elif (LINUX_VERSION_CODE >= 0x020200 && LINUX_VERSION_CODE < 0x020300)
  file->f_dentry->d_inode->i_count++;
  vma->vm_file = file;
#elif LINUX_VERSION_CODE >= 0x020400
  atomic_inc (&file->f_dentry->d_inode->i_count);
  vma->vm_file = file;
#endif

  /* FIXME: where to deassign / decrement? */

  return OK;
}

/*
 * iti_os_select|poll
 */

#if LINUX_VERSION_CODE < 0x020100
static int
iti_os_select (struct inode *inode,
	       struct file *file,
	       int type,
	       select_table *wait)
#elif LINUX_VERSION_CODE >= 0x020200
static unsigned int
iti_os_poll (struct file *file,
	     struct poll_table_struct *wait)
#endif
{
  int error, result, mask, board, funct;
  itifg_t *iti;

#if LINUX_VERSION_CODE > 0x020200
  int type;
#endif

#if LINUX_VERSION_CODE < 0x020100
  if ((error = iti_decode_devno (inode->i_rdev, &board, &funct)))
    return error;
#elif LINUX_VERSION_CODE >= 0x020200
  if ((error = iti_decode_devno (file->f_dentry->d_inode->i_rdev,
				 &board, &funct)))
    return error;
#endif  

  iti = itifg[board];

#if LINUX_VERSION_CODE < 0x020100
  mask = FALSE;
#elif LINUX_VERSION_CODE >= 0x020200
  /* check both in one */
  type = ITI_SEL_IN | ITI_SEL_EX;
  mask = 0;
#endif
  
#if LINUX_VERSION_CODE < 0x020100
  result = iti_select (file, inode->i_rdev, type);
#elif LINUX_VERSION_CODE >= 0x020200
  result = iti_select (file, file->f_dentry->d_inode->i_rdev, type);
#endif

  if (type & ITI_SEL_IN)
    {
      if (result & ITI_SEL_IN)
	{
	  iti->osp.cam_s_wait = FALSE;
#if LINUX_VERSION_CODE < 0x020100
	  mask = TRUE;
#elif LINUX_VERSION_CODE >= 0x020200
	  mask |= (POLLIN | POLLRDNORM);
#endif
	}
      else
	iti->osp.cam_s_wait = TRUE;
    }

  if (type & ITI_SEL_EX)
    if (result & ITI_SEL_EX)
#if LINUX_VERSION_CODE < 0x020100
      mask = TRUE;
#elif LINUX_VERSION_CODE >= 0x020200
      mask |= POLLPRI;
#endif

  if (!mask)
#if LINUX_VERSION_CODE < 0x020100
    select_wait (&iti->osp.cam_s_queue, wait);
#elif LINUX_VERSION_CODE >= 0x020200
    poll_wait (file, &iti->osp.cam_s_queue, wait);
#endif
 
  return mask;
}

/*
 * iti_os_fasync
 */

#if LINUX_VERSION_CODE < 0x020100
static int
iti_os_fasync (struct inode* inode, struct file* file, int mode)
{
  return iti_fasync (file, inode, inode->i_rdev, -1, mode);
}
#elif LINUX_VERSION_CODE >= 0x020200
static int
iti_os_fasync (int desc, struct file *file, int mode)
{
  return iti_fasync (file, file->f_dentry->d_inode,
		     file->f_dentry->d_inode->i_rdev, desc, mode);
}
#endif

/*
 * iti_os_read_proc
 */

#if LINUX_VERSION_CODE < 0x020300
static int
iti_os_read_proc (char *buffer, char **start,
		  off_t offset, int length, int unused)
#elif LINUX_VERSION_CODE >= 0x020400
static int
iti_os_read_proc (char *buffer, char **start,
		  off_t offset, int length, int *eof, void *unused)
#endif
{
  int error, count;
  
  if ((error = iti_os_proc_info (buffer, &count)))
    return error;
  
  return count;
}

/*
 * iti_os_proc_info
 */

int
iti_os_proc_info (char *buffer, int *ret_count)
{
  int count, board , module;
  char mod_name[16];
  
  itifg_t* iti = NULL;

  board = 0;
  ITI_PRINT1("> iti_info\n");

  if (!buffer)
    return ERROR;

  count = 0;

  strcpy (mod_name, "Unknown");
  count += sprintf (&buffer[count], "%s %d.%d.%d-%d\n", PFS_VERSION_IDENT,
		     (iti_version_code & 0xFF000000) >> 24,
		     (iti_version_code & 0x00FF0000) >> 16,
		     (iti_version_code & 0x0000FF00) >> 8,
		     (iti_version_code & 0x000000FF)); 
  count += sprintf (&buffer[count], "%s %s\n", PFS_DATE_IDENT, __DATE__); 
  count += sprintf (&buffer[count], "%s %d\n", PFS_BOARDS_IDENT, boards);
  count += sprintf (&buffer[count], "%s", PFS_MODULES_IDENT);
  for (board = 0; board < boards; board++)
    {
      iti = itifg[board];

      for (module = 0;
	   module < sizeof(mod_names) / sizeof(mod_names[0]);
	   module++)
	if (MODULE_ID == mod_names[module].id)
	  strcpy (mod_name, mod_names[module].name);
      count += sprintf (&buffer[count], " %s", mod_name);
    }
  count += sprintf (&buffer[count], "\n");

  if (ret_count)
    *ret_count = count;

  ITI_PRINT1("< iti_info\n");
  return OK;
}

/*
 * iti_os_init
 */

int
iti_os_init (itifg_t *iti)
{
  ITI_PRINT1("> iti_os_init\n");

  /* timer init */
  init_timer (&iti->osp.cam_timer);
  init_timer (&iti->osp.iop_timer);

#if LINUX_VERSION_CODE >= 0x020200
  iti->osp.before = 0xdeadbeef;
  spin_lock_init (&(iti->osp.spin_lock));
  iti->osp.after = 0xdeadbeef;
#endif

  ITI_PRINT1("< iti_os_init\n");  
  return OK;
}

/*
 * iti_os_probe
 */

static int state = -1;

#if LINUX_VERSION_CODE < 0x020100
#define pci_dev iti_dev
struct pci_dev
{
  u_char bus; 
  u_char dev_fn;
} tmp_pci_dev[ITI_BOARDS_MAX];
static int
iti_os_probe (struct pci_dev **new_dev, struct pci_dev *old_dev)
{
  int error;

  static int iti_board = 0;
  static u_short iti_index = 0;

  ITI_PRINT1("> iti_os_probe\n");

  if (state < 0)
    {
      /* bios installed ? */
      if (!pcibios_present())
	{
	  iti_printw ("no 'pci-bios' present.\n");
	  return -ENXIO;
	}
      state++;
    }
  
  /* search for all known ITI products */
  do
    switch (state)
      {
      case 0:
	ITI_PRINT1("    icpci\n");
	if ((error =
	     pcibios_find_device (ITI_PCI_VENDOR_ID_IMAGINGTECH,
				  ITI_PCI_DEVICE_ID_IMAGINGTECH_ICPCI,
				  iti_index,
				  &tmp_pci_dev[iti_board].bus,
				  &tmp_pci_dev[iti_board].dev_fn))
	    != PCIBIOS_SUCCESSFUL)
	  state++, iti_index = 0;
	else
	  *new_dev = &tmp_pci_dev[iti_board++],
	    iti_index++;
	break;
      case 1:
	ITI_PRINT1("    async\n");
	if ((error =
	     pcibios_find_device (ITI_PCI_VENDOR_ID_IMAGINGTECH,
				  ITI_PCI_DEVICE_ID_IMAGINGTECH_ASYNC,
				  iti_index,
				  &tmp_pci_dev[iti_board].bus,
				  &tmp_pci_dev[iti_board].dev_fn))
	    != PCIBIOS_SUCCESSFUL)
	  state++, iti_index = 0;
	else
	  *new_dev = &tmp_pci_dev[iti_board++],
	    iti_index++;
	break;
      case 2:
	ITI_PRINT1("    pcdig\n");
	if ((error =
	     pcibios_find_device (ITI_PCI_VENDOR_ID_IMAGINGTECH,
				  ITI_PCI_DEVICE_ID_IMAGINGTECH_PCDIG,
				  iti_index,
				  &tmp_pci_dev[iti_board].bus,
				  &tmp_pci_dev[iti_board].dev_fn))
	    != PCIBIOS_SUCCESSFUL)
	  state++, iti_index = 0;
	else
	  *new_dev = &tmp_pci_dev[iti_board++],
	    iti_index++;
	break;
      default:
	ITI_PRINT1("    none\n");
	return -ENXIO;
      }
  while (error != PCIBIOS_SUCCESSFUL);

  ITI_PRINT1("< iti_os_probe\n");
  
  return OK;
}
#endif
#if LINUX_VERSION_CODE >= 0x020200
static int
iti_os_probe (struct pci_dev **new_dev, struct pci_dev *old_dev)
{
  struct pci_dev *tmp_dev = old_dev;

  ITI_PRINT1("> iti_os_probe\n");

  if (state < 0)
    {
      /* is a pci bios installed and supported */
      if (!pci_present())
	{
	  iti_printw ("no 'pcibios' present.\n");
	  return -ENXIO;
	}
      state++;
    }
  
  /* search for all known ITI products */
  do
    switch (state)
      {
      case 0:
	ITI_PRINT1("    icpci\n");
	if (!(*new_dev =
	      pci_find_device (ITI_PCI_VENDOR_ID_IMAGINGTECH,
			       ITI_PCI_DEVICE_ID_IMAGINGTECH_ICPCI,
			       tmp_dev)))
	  {
	    state++;
	    tmp_dev = NULL;
	  }
	break;
      case 1:
	ITI_PRINT1("    async\n");
	if (!(*new_dev =
	      pci_find_device (ITI_PCI_VENDOR_ID_IMAGINGTECH,
			       ITI_PCI_DEVICE_ID_IMAGINGTECH_ASYNC,
			       tmp_dev)))
	  {
	    state++;
	    tmp_dev = NULL;
	  }
	break;
      case 2:
	ITI_PRINT1("    pcdig\n");
	if (!(*new_dev = 
	      pci_find_device (ITI_PCI_VENDOR_ID_IMAGINGTECH,
			       ITI_PCI_DEVICE_ID_IMAGINGTECH_PCDIG,
			       tmp_dev)))
	  {
	    state++;
	    tmp_dev = NULL;
	  }
	break;
      default:
	ITI_PRINT1("    none\n");
	return -ENXIO;
      }
  while (*new_dev == NULL);
  
  ITI_PRINT1("< iti_os_probe\n");
  return OK;
}
#endif

/*
 * iti_os_pci_init
 */

#if LINUX_VERSION_CODE < 0x020100
int
iti_os_pci_init (itifg_t *iti, void *data)
{
  int error;

  u_char iti_bus, iti_dev_fn, irq_line, lat_tim, rev_id, lev_id;

  u_short iti_command, device;

  u_long p_icr_base, p_cr_base,	p_mod_base, p_mem_base;

  struct pci_dev *iti_dev = (struct pci_dev*)data;

  ITI_PRINT1("> iti_os_pci_init\n");

  /* assign device ident for further calls */
  iti_bus = iti_dev->bus;
  iti_dev_fn = iti_dev->dev_fn;
  
  /* read device id */
  error = pcibios_read_config_word (iti_bus, iti_dev_fn,
				    PCI_DEVICE_ID,
				    &device);
  if (error != PCIBIOS_SUCCESSFUL) {
    iti_printw ("cannot read device id.\n");
    return (-ENXIO);
  }

  /* read revision id */
  error = pcibios_read_config_byte (iti_bus, iti_dev_fn,
				    PCI_REVISION_ID,
				    &rev_id);
  if (error != PCIBIOS_SUCCESSFUL) {
    iti_printw ("cannot read revision id.\n");
    return (-ENXIO);
  }

  /* read revision level id from status register (??)*/
  error = pcibios_read_config_byte (iti_bus, iti_dev_fn,
				    PCI_STATUS,
				    &lev_id);
  if (error != PCIBIOS_SUCCESSFUL) {
    iti_printw ("cannot read status register.\n");
    return (-ENXIO);
  }
    
  /* base address of interface control register set */
  error = pcibios_read_config_dword (iti_bus, iti_dev_fn,
				     PCI_BASE_ADDRESS_0,
				     (unsigned int*)&p_icr_base);
  if (error != PCIBIOS_SUCCESSFUL) {
    iti_printw ("cannot read base address 0.\n");
    return (-EIO);
  }

  /* base address of control register set */
  error = pcibios_read_config_dword (iti_bus, iti_dev_fn,
				     PCI_BASE_ADDRESS_1,
				     (unsigned int*)&p_cr_base);
  if (error != PCIBIOS_SUCCESSFUL) {
    iti_printw ("cannot read base address 1.\n");
    return (-ENXIO);
  }

  /* base address of acquisition module register set OR
       base address of frame grabber on board memory */
  error = pcibios_read_config_dword (iti_bus, iti_dev_fn,
				     PCI_BASE_ADDRESS_2,
				     (unsigned int*)&p_mod_base);
  if (error != PCIBIOS_SUCCESSFUL) {
    iti_printw ("cannot read base address 2.\n");
    return (-ENXIO);
  }

  /* base address of frame grabber on board memory OR unused */
  error = pcibios_read_config_dword (iti_bus, iti_dev_fn,
				     PCI_BASE_ADDRESS_3,
				     (unsigned int*)&p_mem_base);
  if (error != PCIBIOS_SUCCESSFUL) {
    iti_printw ("cannot read base address 3.\n");
    return (-ENXIO);
  }

  /* interrupt line */
  error = pcibios_read_config_byte (iti_bus, iti_dev_fn,
				    PCI_INTERRUPT_LINE,
				    &irq_line);
  if (error != PCIBIOS_SUCCESSFUL) {
    iti_printw ("cannot read interrupt line.\n");
    return (-ENXIO);
  }
    
  /* latency timer (if too low -> correct) */
  error = pcibios_read_config_byte (iti_bus, iti_dev_fn,
				    PCI_LATENCY_TIMER,
				    &lat_tim);
  if (error != PCIBIOS_SUCCESSFUL) {
    iti_printw ("cannot read latency timer.\n");
    return (-ENXIO);
  }
  if (lat_tim < ITI_LATENCY_DEF) {
    error = pcibios_write_config_byte (iti_bus, iti_dev_fn,
				       PCI_LATENCY_TIMER,
				       ITI_LATENCY_DEF);
    if (error != PCIBIOS_SUCCESSFUL) {
      iti_printw ("cannot write latency timer.\n");
      return (-ENXIO);
    }
  }

  /* command register (memory access, i/o access, bus master mode) */
  error = pcibios_read_config_word (iti_bus, iti_dev_fn,
				    PCI_COMMAND,
				    &iti_command);
  if (error != PCIBIOS_SUCCESSFUL) {
    iti_printw ("cannot read command register.\n");
    return (-ENXIO);
  }
  iti_command |= (PCI_COMMAND_IO |
		  PCI_COMMAND_MEMORY |
		  PCI_COMMAND_MASTER);
  error = pcibios_write_config_word (iti_bus, iti_dev_fn,
				     PCI_COMMAND,
				     iti_command);
  if (error != PCIBIOS_SUCCESSFUL) {
    iti_printw ("cannot write command register.\n");
    return (-ENXIO);
  }

  /* debugging */
  ITI_PRINT1A("    Bus       : %d\n", iti_bus);
  ITI_PRINT1A("    Device    : %d\n", PCI_SLOT(iti_dev_fn));
  ITI_PRINT1A("    Function  : %d\n", PCI_FUNC(iti_dev_fn));
  ITI_PRINT1A("    Latency   : %d\n", lat_tim);

  /* fill in structure */
  iti->address[0] = p_icr_base;
  iti->address[1] = p_cr_base;
  iti->address[2] = p_mod_base;
  iti->address[3] = p_mem_base;
  iti->address[4] = 0;
  iti->address[5] = 0;
  iti->irq_line = irq_line;
  iti->revision = rev_id;
  iti->status = lev_id;
  iti->device_id = device;

  ITI_PRINT1("< iti_os_pci_init\n");
  return OK;
}
#endif
#if LINUX_VERSION_CODE >= 0x020200
int
iti_os_pci_init (itifg_t *iti, void *data)
{
  u_char iti_bus;
  u_int iti_devfn;
  
  struct pci_dev *iti_dev = (struct pci_dev*)data;

  ITI_PRINT1("> iti_os_pci_init\n");

  /* assign device ident for further calls */
  iti_bus = iti_dev->bus->number;
  iti_devfn = iti_dev->devfn;
        
  /* read device id */
  iti->device_id = iti_dev->device;
  
  /* read status register */
  if (pcibios_read_config_word (iti_bus, iti_devfn,
				PCI_STATUS,
				&iti->status))
    {
      iti_printw ("can't read status register.\n");
      return -EIO;
    }
  
  /* read revision id */
  if (pcibios_read_config_byte (iti_bus, iti_devfn,
				PCI_REVISION_ID,
				&iti->revision))
    {
      iti_printw ("can't read revision id.\n");
      return -EIO;
    }

  /* base addresses of various regions */
#if LINUX_VERSION_CODE < 0x020300
  iti_memcpy (iti->address, iti_dev->base_address,
	      sizeof(iti->address));
#elif LINUX_VERSION_CODE >= 0x020400
  {
    int count;
    for (count = 0; count < 6; count++)
      iti->address[count] = iti_dev->resource[count].start;
  } 
#endif
  
  /* interrupt line */
  iti->irq_line = iti_dev->irq;

  /* enable bus master mode and adjust latency timer */
  pci_set_master (iti_dev);
  
  ITI_PRINT1A("    Bus       : %d\n", iti_bus);
  ITI_PRINT1A("    Device    : %d\n", PCI_SLOT(iti_devfn));
  ITI_PRINT1A("    Function  : %d\n", PCI_FUNC(iti_devfn));
  
  ITI_PRINT1("< iti_os_pci_init\n");
  return OK;
}
#endif

/*
 * iti_os_map_init
 */

#if LINUX_VERSION_CODE < 0x020100
int
iti_os_map_init (itifg_t *iti, void *data,
		 u_long *align_div, long *align_mod)
{
  void *v_icr_base = NULL, *v_cr_base = NULL;
  void *v_mod_base = NULL, *v_mem_base = NULL;

  ITI_PRINT1("> iti_os_map_init\n");

  /* remap all register spaces and on board memory to low kernel memory */
  if (!iti->regs_io)
    {
      if ((v_icr_base = vremap (align_div[ICR_REGS],
				ROUND_UPTO_PAGE(align_mod[ICR_REGS] + 
						ICR_REG_SIZE))) == NULL)
	{
	  iti_printw ("unable to ioremap interface control regs.\n");
	  return -EFAULT;
	}
      iti->v_icr_base = v_icr_base;
    }
  
  switch (DEVICE_ID)
    {
    case ITI_PCI_DEVICE_ID_IMAGINGTECH_ICPCI:
      if (!iti->regs_io)
	{
	  if ((v_cr_base = vremap (align_div[CR_REGS],
				   ROUND_UPTO_PAGE(align_mod[CR_REGS] + 
						   ICP_CR_REG_SIZE))) == NULL) 
	    {
	      iti_printw ("unable to ioremap control regs.\n");
	      return -EFAULT;
	    }
	  iti->v_cr_base = v_cr_base;
	}
      if ((v_mod_base = vremap (iti->address[2], ICP_MOD_MAP_SIZE)) == NULL)
	{
	  iti_printw ("unable to ioremap acquisition modul regs.\n");
	  return -EFAULT;
	}
      iti->v_mod_base = v_mod_base;
      if (iti->big_mem)
	{
	  if ((v_mem_base = vremap (iti->address[3],
				    ICP_MEM_MAP_SIZE << 1)) == NULL)
	    {
	      iti_printw ("unable to ioremap main board memory.\n");
	      return -EFAULT;
	    }
	}
      else
	{
	  if ((v_mem_base = vremap (iti->address[3],
				    ICP_MEM_MAP_SIZE)) == NULL)
	    {
	      iti_printw ("unable to ioremap main board memory.\n");
	      return -EFAULT;
	    }
	}
      iti->v_mem_base = v_mem_base;
      break;
      
    case ITI_PCI_DEVICE_ID_IMAGINGTECH_PCDIG:
      if ((v_cr_base = vremap (align_div[CR_REGS], PCD_CR_REG_SIZE)) == NULL)
	{
	  iti_printw ("unable to ioremap control regs.\n");
	  return -EFAULT;
	}
      iti->v_cr_base = v_cr_base;
      /* not used at the PCDig */
      iti->v_mod_base = NULL;
      if ((v_mem_base = vremap (iti->address[2], PCD_MEM_MAP_SIZE)) == NULL)
	{
	  iti_printw ("unable to ioremap main board memory.\n");
	  return (-EFAULT);
	}
      iti->v_mem_base = v_mem_base;
      break;
      
    default:
      return -ENXIO;
    }
  
  ITI_PRINT1("< iti_os_map_init\n");    
  return OK;
}
#elif LINUX_VERSION_CODE >= 0x020200
int
iti_os_map_init (itifg_t *iti, void *data,
		 u_long *align_div, long *align_mod)
{
  ITI_PRINT1("> iti_os_map_init\n");

  /* remap all register spaces to low kernel memory */
  if (!iti->regs_io)
    {
      iti->v_icr_base =
	ioremap_nocache (align_div[ICR_REGS],
			 ROUND_UPTO_PAGE(align_mod[ICR_REGS] +
					 ICR_REG_SIZE));
      if (iti->v_icr_base == NULL)
	{
	  iti_printw ("unable to ioremap interface control regs.\n");
	  return -EFAULT;
	}
    }
  
  switch (iti->device_id)
    {
    case ITI_PCI_DEVICE_ID_IMAGINGTECH_ICPCI:
      if (!iti->regs_io)
	{
	  iti->v_cr_base =
	    ioremap_nocache (align_div[CR_REGS],
			     ROUND_UPTO_PAGE(align_mod[CR_REGS] +
					     ICP_CR_REG_SIZE));
	  if (iti->v_cr_base == NULL)
	    {
	      iti_printw ("unable to ioremap control regs.\n");
	      return -EFAULT;
	    }
	}
      iti->v_mod_base = ioremap_nocache (iti->address[2],
					 ROUND_UPTO_PAGE(ICP_MOD_MAP_SIZE));
      if (iti->v_mod_base == NULL)
	{
	  iti_printw ("unable to ioremap acquisition modul regs.\n");
	  return -EFAULT;
	}
      if (iti->big_mem)
	iti->v_mem_base = ioremap (iti->address[3],
				   ROUND_UPTO_PAGE(ICP_MEM_MAP_SIZE << 1));
      else
	iti->v_mem_base = ioremap (iti->address[3],
				   ROUND_UPTO_PAGE(ICP_MEM_MAP_SIZE));
      if(iti->v_mem_base == NULL)
	{
	  iti_printw ("unable to ioremap main board memory.\n");
	  return -EFAULT;
	}
      break;

    case ITI_PCI_DEVICE_ID_IMAGINGTECH_PCDIG:
      iti->v_cr_base = ioremap_nocache (align_div[CR_REGS], PCD_CR_REG_SIZE);
      if (iti->v_cr_base == NULL)
	{
	  iti_printw ("unable to ioremap control regs.\n");
	  return -EFAULT;
	}
      /* not used at the PCDig */
      iti->v_mod_base = NULL;
      iti->v_mem_base =	ioremap (iti->address[2],
				 ROUND_UPTO_PAGE(PCD_MEM_MAP_SIZE));
      if(iti->v_mem_base == NULL)
	{
	  iti_printw ("unable to ioremap main board memory.\n");
	  return -EFAULT;
	}
      break;

    default:
      return -ENXIO;
    }

  ITI_PRINT1("< iti_os_map_init\n");
  return OK;
}
#endif

#ifdef MODULE
/*
 * iti_os_detach
 */

#if LINUX_VERSION_CODE < 0x020100
#define iounmap vfree
#endif

void
iti_os_detach (void)
{  
  int board;
  itifg_t *iti;

  ITI_PRINT1("> iti_os_detach\n");
  
  board = 0;
  while (itifg[board] != NULL)
    {
      iti = itifg[board];

      iti_free_board (iti);
      if (iti->v_mem_base != NULL)
	{
	  ITI_PRINT3A("    iti.map->v_mem_base: 0x%p\n", iti->v_mem_base);
	  iounmap (iti->v_mem_base);
	}
      if (iti->v_mod_base != NULL)
	{
	  ITI_PRINT3A("    iti.map->v_mod_base: 0x%p\n",iti->v_mod_base);
	  iounmap (iti->v_mod_base);
	}
      if (iti->v_cr_base != NULL)
	{
	  iti->v_cr_base =
	    (u_char*)ROUND_DOWNTO_PAGE(iti->v_cr_base);
	  ITI_PRINT3A("    iti.map->v_cr_base: 0x%p\n", iti->v_cr_base);
	  iounmap (iti->v_cr_base);
	}
      if (iti->v_icr_base != NULL)
	{
	  iti->v_icr_base = (u_long*)ROUND_DOWNTO_PAGE(iti->v_icr_base);
	  ITI_PRINT3A("    iti.map->v_icr_base: 0x%p\n", iti->v_icr_base);
	  iounmap (iti->v_icr_base);
	}
      free_irq (iti->irq_line, iti);

      board++;
    }

#if LINUX_VERSION_CODE < 0x020300
  proc_unregister (&proc_root, iti_proc_entry.low_ino);
#elif LINUX_VERSION_CODE >= 0x020400
  remove_proc_entry(PFS_STRING, NULL);
#endif

  if (unregister_chrdev(major, DEV_STRING) < 0)
    iti_printwa ("error while unregistering %s (%d).\n", DEV_STRING, major);
  
  iti_detach ();
  
  ITI_PRINT1("< iti_os_detach\n");
}

/*
 * iti_os_attach
 */

int
iti_os_attach (void)
{
  int error, board;
  itifg_t *iti;

  struct pci_dev *iti_dev[ITI_BOARDS_MAX];

  ITI_PRINT1("> iti_os_attach\n");

  iti_memset (iti_dev, 0, ITI_BOARDS_MAX * sizeof(struct pci_dev*));
  
  /* probe pci configuration space for possible devices */
  boards = 0;
  while (!iti_os_probe (&iti_dev[boards],
			boards ? iti_dev[boards - 1] : NULL))
    boards++;
  
  if (boards)
    iti_printia ("%d device(s) found.\n", boards);
  else
    {
      iti_printi ("no devices found.\n");
      return OK;
    }

  /* how are the major devno's assigned (default is 0 - dynamic) */
  if (major != 0)
    {
      /* non automatic */
      if (major > MAX_CHRDEV)
	{
	  iti_printwa ("major device number %d out of range (1-%d).\n",
		       major, MAX_CHRDEV);
	  iti_printwa ("using device number %d instead.\n", ITI_MAJOR);
	  major = ITI_MAJOR;	/* old driver behavior */
	}
      /* register first character device */
      if ((error = register_chrdev (major, DEV_STRING, &iti_fops)) < 0)
      {
	iti_printwa ("'register_chrdev (character device %d)' failed.\n",
		     major);
	cleanup_module();
	return error;
      }
    }
  else
    {
      /* register first character device */
      if ((major = register_chrdev (major, DEV_STRING, &iti_fops)) < 0)
	{
	  iti_printwa ("'register_chrdev (character device %d)' failed.\n",
		       major);
	  cleanup_module();
	  return major;
	}
    }  

  iti_memset (itifg, 0, ITI_BOARDS_MAX * sizeof(itifg_t*));

  /* init all driver instances */
  for (board = 0; board < boards; board++)
    {
      if ((error = iti_attach (board, (void*)iti_dev[board])))
	{
	  iti_os_detach ();
	  return error;
	}
      iti = itifg[board];
      /* register per board interrupts */
      if (iti->irq_line != 255)
	if ((error = request_irq (iti->irq_line,
				  &iti_os_interrupt,
				  SA_SHIRQ,
				  IRQ_STRING, iti)))
	  if ((error = request_irq (iti->irq_line,
				    &iti_os_interrupt,
				    SA_INTERRUPT | SA_SHIRQ,
				    IRQ_STRING, iti)))
	    {
	      iti_printwa ("'request_irq (intr. %d)' failed.\n",
			   iti->irq_line);
	      iti_os_detach ();
	      return error;
	    }
    }
  
#if LINUX_VERSION_CODE < 0x020100  
  proc_register_dynamic (&proc_root, &iti_proc_entry);
#elif (LINUX_VERSION_CODE >= 0x020200 && LINUX_VERSION_CODE < 0x020300)
  proc_register (&proc_root, &iti_proc_entry);
#elif LINUX_VERSION_CODE >= 0x020400
  iti_proc_entry = create_proc_entry(PFS_STRING, 0, 0);
  if (!iti_proc_entry)
    iti_printw ("register_proc_entry failed.\n");
  else
    iti_proc_entry->read_proc = iti_os_read_proc;
#endif

  /* print version */
  iti_printia ("version %d.%d.%d-%d.\n",
	       (iti_version_code & 0xFF000000) >> 24,
	       (iti_version_code & 0x00FF0000) >> 16,
	       (iti_version_code & 0x0000FF00) >> 8,
	       (iti_version_code & 0x000000FF));
  iti_printia ("%s (linux %s).\n", __DATE__, OS_RELEASE);

#ifdef DEBUG_ITIFG
  iti_printia ("Debug output enabled, debug level is %d (%x hex)\n",
	       DEBUG_ITIFG, DEBUG_ITIFG);
#endif
  
  ITI_PRINT1("< iti_os_attach\n");  
  return OK;
}
#endif /* MODULE */

/* these functions are only called, if NOT using the inline equivalents */

u_short
iti_readw (u_short *addr) 
{
  return readw (addr);
}
void
iti_writew (u_short value, u_short *addr)
{
  writew (value, addr);
}
u_long
iti_readl (u_long *addr) 
{
  return readl (addr);
}
void
iti_writel (u_long value, u_long *addr)
{
  writel (value, addr);
}
u_short
iti_inw (u_short port) 
{
  return inw (port);
}
void
iti_outw (u_short value, u_short port)
{
  outw (value, port);
}
u_long
iti_inl (u_short port) 
{
  return inl (port);
}
void
iti_outl (u_long value, u_short port)
{
  outl (value, port);
}
void*
iti_memset_io (void *dst, int val, size_t size)
{
#if defined(__alpha__) || defined (__ia64__)
  memset_io (dst, val, size);
  return NULL;
#else
  return memset_io (dst, val, size);
#endif
}
void
iti_udelay (u_long time)
{
  udelay (time);
}

#if LINUX_VERSION_CODE < 0x020100
int
iti_copyout (void *dst, void *src, size_t size)
{
  memcpy_tofs (dst, src, size);
  return 0;
}
int
iti_copyin (void *dst, const void *src, size_t size)
{
  memcpy_fromfs (dst, src, size);
  return 0;
}
#elif LINUX_VERSION_CODE >= 0x020200
int
iti_copyout (void *dst, void *src, size_t size)
{
  if (copy_to_user (dst, src, size))
    return -EFAULT;

  return 0;
}
int
iti_copyin (void *dst, const void *src, size_t size)
{
  if (copy_from_user (dst, src, size))
    return -EFAULT;

  return 0;
}
#endif

void*
iti_memcpy (void* dst, const void* src, size_t size)
{
  return memcpy (dst, src, size);
}
void*
iti_memset (void* dst, int value, size_t size)
{
  return memset (dst, value, size);
}
char*
iti_strcpy (char* dst, const char* src)
{
  return strcpy (dst, src);
}
int
iti_strcmp (const char* dst, const char* src)
{
  return strcmp (dst, src);
}

